"""Data object storing the solver parameters at a given datetime for the Simulation runs."""
from dataclasses import dataclass
from abc import ABC


@dataclass
class SolverParameter(ABC):
    """Abstract Data object storing the solver parameters at a given datetime for the Simulation runs.
    Stored as a datetime dependent object as some sims can change these parameters with time.
    """
    date: str



from functools import partial
from typing import Optional, Union
import re

from ResSimpy.Grid import GridArrayDefinition


def strip_file_of_comments(file_as_list: list[str], strip_str: bool = False,
                           comment_characters: Optional[list[str]] = None,
                           square_bracket_comments: bool = False) -> list[str]:
    """Strips all of the inline, single and multi line comments out of a file.
    Comment characters assumed are: ! and square brackets. Escaped characters are ones wrapped in quotation marks.

    Args:
        square_bracket_comments (bool): whether to also remove text contained in square brackets ([])
        comment_characters (Optional[list[str]]): A list of characters that are considered comments. Defaults to the
                                                  Nexus format (!)
        file_as_list (list[str]): a list of strings containing each line of the file as a new entry
        strip_str (bool, optional): if True strips the lines of whitespace. Defaults to False.

    Returns:
        list[str]: a list of strings containing each line of the file as a new entry without comments
    """
    if comment_characters is None:
        comment_characters = ['!']
    file_as_list = list(filter(None, file_as_list))

    file_without_comments = file_as_list

    for comment_character in comment_characters:
        if comment_character == 'C':  # handle VIP comment with single C character at the start of a line
            file_without_comments = [line for line in file_without_comments if not line.startswith('C ') and not
                                     line.strip() == 'C']
        else:
            # remove any empty lines
            # regex: look back and forward 1 character from an ! and check if it's a quotation mark and
            # exclude it from the match if it is
            file_without_comments = [re.split(fr'(?<!\"){comment_character}(?!\")', x)[0]
                                     for x in file_without_comments if x and x[0] != comment_character]

    flat_file = '\n'.join(file_without_comments)

    if square_bracket_comments:
        # regex: look back and forward 1 character from a square bracket and check if it's a quotation mark and
        # exclude it from the match if it is
        flatfile_minus_square_brackets = flat_file
        for comment_character in comment_characters:
            flatfile_minus_square_brackets = re.sub(fr"(?<!\")\[.*?](?{comment_character}\")", '',
                                                    flatfile_minus_square_brackets, flags=re.DOTALL)

        file_without_comments = flatfile_minus_square_brackets.splitlines()

    if strip_str:
        file_without_comments = [x.strip() for x in file_without_comments]
    return file_without_comments


def load_file_as_list(file_path: str, strip_comments: bool = False, strip_str: bool = False) -> list[str]:
    """Reads the text file into a variable.

    Args:
        file_path (str): string containing a path pointing towards a text file
        strip_comments (bool, optional): If set to True removes all inline/single line comments from \
            the passed in file. Defaults to False.
        strip_str (bool, optional): if True strips the lines of whitespace. Defaults to False.

    Returns:
        list[str]: list of strings with each line from the file a new entry in the list
    """
    try:
        with open(file_path, 'r') as f:
            file_content = list(f)
    except UnicodeDecodeError:
        with open(file_path, 'r', errors='replace') as f:
            file_content = list(f)

    if strip_comments:
        file_content = strip_file_of_comments(file_content, strip_str=strip_str)

    return file_content


def get_next_value(start_line_index: int, file_as_list: list[str], search_string: None | str = None,
                   ignore_values: None | list[str] = None,
                   replace_with: str | GridArrayDefinition | None = None,
                   comment_characters: None | list[str] = None,
                   single_c_acts_as_comment: bool = True) -> Optional[str]:
    """Gets the next non blank value in a list of lines.

    Args:
        start_line_index (int): line number to start reading file_as_list from
        file_as_list (list[str]): a list of strings containing each line of the file as a new entry
        search_string (str): string to search from within the first indexed line
        ignore_values (Optional[list[str]], optional): a list of values that should be ignored if found. \
            Defaults to None.
        replace_with (str | VariableEntry | None], optional): a value to replace the existing value with. \
            Defaults to None.
        comment_characters (Optional[list[str]], optional): a list of characters that are considered inline comments.
            Defaults to the Nexus format (!)
        single_c_acts_as_comment: (bool): whether a single C character at the start of a line should be treated as a
            comment. Defaults to Nexus setting which is True.

    Returns:
        Optional[str]: Next non blank value from the list, if none found returns None
    """
    if comment_characters is None:
        comment_characters = ['!']
    invalid_characters = ["\n", "\t", " ", ","]
    invalid_characters.extend(comment_characters)

    value_found = False
    value = ''
    if search_string is None:
        search_string = file_as_list[start_line_index]
    line_index = start_line_index
    while value_found is False and line_index <= len(file_as_list):
        character_location = 0
        new_search_string = False
        line_already_skipped = False

        # If the string is wrapped in quotation marks, return the full string (including invalid characters)
        if len(search_string) > 2 and search_string.startswith("\"") and search_string.endswith("\""):
            value += search_string[1:len(search_string) - 1]
            value_found = True
            break

        for character in search_string:
            # move lines once we hit a comment character or new line character,or are at the end of search string
            starts_with_c_only = (single_c_acts_as_comment and
                                  (character_location == 0 and character == "C" and
                                   (len(search_string) == 1 or search_string[character_location + 1] == ' ')))
            if character in comment_characters or character == "\n" or starts_with_c_only:
                line_index += 1
                line_already_skipped = True
                # If we've reached the end of the file, return None
                if line_index >= len(file_as_list):
                    return None
                # Move to the next line down in file_as_list
                temp_search_string = file_as_list[line_index]
                if not isinstance(temp_search_string, str):
                    raise ValueError(f'No valid value found, hit INCLUDE statement instead on line number \
                        {line_index}')
                search_string = temp_search_string
                break
            elif character not in invalid_characters:
                character_location, new_search_string, search_string, value = (
                    __extract_substring_until_next_invalid_character(character_location, ignore_values,
                                                                     invalid_characters, new_search_string,
                                                                     search_string, value))
                # if the substring is a not an empty string, we've found a value
                if value != "":
                    value_found = True
                    # Replace the original value with the new requested value
                    if replace_with is not None:
                        value = __replace_value_in_file_as_list(file_as_list, line_index, replace_with, value)
                    break

            if new_search_string:
                break

            character_location += 1
        if not line_already_skipped:
            line_index += 1
        if line_index <= len(file_as_list) - 1:
            search_string = file_as_list[line_index]

    if not value_found:
        return None

    return value


def __replace_value_in_file_as_list(file_as_list: list[str], line_index: int, replace_with: str | GridArrayDefinition,
                                    value: str) -> str:
    """Replaces a value in a file_as_list with a new value.

    Args:
        file_as_list (list[str]): a list of strings containing each line of the file as a new entry
        line_index (int): line number from file_as_list to replace the value in.
        replace_with (str | VariableEntry): a value to replace the existing value with.
        value (str): the value to be replaced.

    Returns:
        str: the value that has been replaced.
    """
    original_line = file_as_list[line_index]
    if not isinstance(original_line, str):
        raise ValueError(f'No valid value found, hit INCLUDE statement instead on line number \
                                            {line_index}')
    new_line = original_line
    if isinstance(replace_with, str):
        new_value = replace_with
    elif isinstance(replace_with, GridArrayDefinition):
        new_line, new_value, value = __replace_with_variable_entry(new_line, original_line,
                                                                   replace_with, value)
    if new_value is None:
        raise ValueError(f'Value for replacing has returned a null value,\
                            check replace_with input, {replace_with=}')
    new_line = new_line.replace(value, new_value, 1)
    file_as_list[line_index] = new_line
    return value


def __extract_substring_until_next_invalid_character(character_location: int,
                                                     ignore_values: list[str] | None,
                                                     invalid_characters: list[str],
                                                     new_search_string: bool,
                                                     search_string: str,
                                                     value: str) -> tuple[int, bool, str, str]:
    """Extracts a substring from the search_string until an invalid character is found.

    Args:
        character_location (int): current location in the search_string
        ignore_values (list[str] | None): list of values to ignore if found
        invalid_characters (list[str]): list of characters that are considered invalid
        new_search_string (bool): whether we are starting a new search string
        search_string (str): string to search from within the first indexed line
        value (str): current value being built up
    Returns:
        int: new character location in the search_string
        bool: whether we are starting a new search string
        str: new search string if a value has been ignored
        str: the value that has been built up, or blank if a value has been ignored
    """
    value_string = search_string[character_location: len(search_string)]
    if ignore_values is not None:
        ignore_values = [x.upper() for x in ignore_values]
    confirm_exclude: bool = False

    for value_character in value_string:
        if value_character not in invalid_characters:
            value += value_character

        if confirm_exclude and value_character in invalid_characters:
            # String to ignore found, ignore it and get the next value
            search_string = search_string[character_location: len(search_string)]
            new_search_string = True
            value = ""

        confirm_exclude = False

        # If we've formed a string value we're supposed to ignore, check the next character ends the value on the next
        # loop through. Avoids us excluding values that happen to occur in a longer string.
        if ignore_values is not None and value.upper() in ignore_values:
            if character_location >= len(value_string) - 1:
                value = ""
                break
            else:
                confirm_exclude = True

        character_location += 1

        # stop adding to the value once we hit an invalid_character
        if value_character in invalid_characters and value != '':
            break
    return character_location, new_search_string, search_string, value


def __replace_with_variable_entry(new_line: str, original_line: str, replace_with: GridArrayDefinition, value: str)\
        -> tuple[str, str, str]:
    new_value = replace_with.value if replace_with.value is not None else ''
    if replace_with.modifier != 'VALUE':
        new_line = new_line.replace('INCLUDE ', '')
    elif 'INCLUDE' not in original_line:
        new_value = 'INCLUDE ' + replace_with.value if replace_with.value is not None else ''
    # If we are replacing the first value from a mult, remove the space as well
    if replace_with.modifier == 'MULT' and replace_with.value == '':
        value += ' '
    return new_line, new_value, value


def check_token(token: str, line: str) -> bool:
    """Checks if the text line contains the supplied token and is not commented out
    Args:
        token (str): keyword value to search the line for
        line (str): string to search the token in
    Returns:
        bool: True if the text line contains the supplied token, False otherwise.
    """
    token_separator_chars = [" ", '\n', '\t', '!']
    uppercase_line = line.upper()
    token_location = uppercase_line.find(token.upper())

    # Not found at all, return false
    if token_location == -1:
        return False

    if line.startswith('C '):
        return False

    comment_character_location = line.find("!")

    # Check if the line is commented out before the token appears
    if comment_character_location != -1 and comment_character_location < token_location:
        return False

    # Check if the character before the token is a separator such as space, tab etc and not another alphanumeric char
    if token_location != 0 and line[token_location - 1] not in token_separator_chars:
        return False

    # Check if the character after the token is a separator such as space, tab etc and not another alphanumeric char
    if token_location + len(token) != len(line) and line[token_location + len(token)] not in token_separator_chars:
        return False

    return True


def value_in_file(token: str, file: list[str]) -> bool:
    """Returns true if a token is found in the specified file.

    Args:
        token (str): the token being searched for.
        file (list[str]): a list of strings containing each line of the file as a new entry

    Returns:
        bool: True if the token is found and False otherwise
    """
    token_found = any(map(partial(check_token, token), file))

    return token_found


def get_token_value(token: str, token_line: str, file_list: list[str],
                    ignore_values: Optional[list[str]] = None,
                    replace_with: Union[str, GridArrayDefinition, None] = None) -> Optional[str]:
    """Gets the value following a token if supplied with a line containing the token.

    Arguments:
        token (str): the token being searched for.
        token_line (str): string value of the line that the token was found in.
        file_list (list[str]): a list of strings containing each line of the file as a new entry
        ignore_values (list[str], optional): a list of values that should be ignored if found. \
            Defaults to None.
        replace_with (Union[str, VariableEntry, None], optional):  a value to replace the existing value with. \
            Defaults to None.

    Returns:
        Optional[str]: The value following the supplied token, if it is present.
    """

    token_upper = token.upper()
    token_line_upper = token_line.upper()

    # If this line is commented out, don't return a value
    if "!" in token_line_upper and token_line_upper.index("!") < token_line_upper.index(token_upper):
        return None

    search_start = token_line_upper.index(token_upper) + len(token) + 1
    search_string = token_line[search_start: len(token_line)]
    line_index = file_list.index(token_line)

    # If we have reached the end of the line, go to the next line to start our search
    if len(search_string) < 1:
        line_index += 1
        if line_index >= len(file_list):
            return None
        search_string = file_list[line_index]
    if not isinstance(search_string, str):
        raise ValueError
    value = get_next_value(line_index, file_list, search_string, ignore_values, replace_with)
    return value


def get_expected_token_value(token: str, token_line: str, file_list: list[str],
                             ignore_values: Optional[list[str]] = None,
                             replace_with: Union[str, GridArrayDefinition, None] = None,
                             custom_message: Optional[str] = None) -> str:
    """Function that returns the result of get_token_value if a value is found, otherwise it raises a ValueError.

    Args:
        token (str): the token being searched for.
        token_line (str): string value of the line that the token was found in.
        file_list (list[str]): a list of strings containing each line of the file as a new entry
        ignore_values (list[str], optional): a list of values that should be ignored if found. \
            Defaults to None.
        replace_with (Union[str, VariableEntry, None], optional):  a value to replace the existing value with. \
            Defaults to None.
        custom_message Optional[str]: A custom error message if no value is found.

    Returns:
        str:  The value following the supplied token, if it is present.

    Raises:
        ValueError if a value is not found
    """
    value = get_token_value(token, token_line, file_list, ignore_values, replace_with)

    if value is None:
        if custom_message is None:
            raise ValueError(f"No value found in the line after the expected token ({token}), line: {token_line}")
        else:
            raise ValueError(f"{custom_message} {token_line}")

    return value


def load_in_three_part_date(initial_token: str, token_line: str, file_as_list: list[str], start_index: int) -> str:
    """Function that reads in a three part date separated by spaces e.g. 1 JAN 2024.

    Args:
    initial_token (str): The token that will appear before the start of the date e.g. DATE
    token_line (str): Line in the file that the token has been found.
    file_as_list (list[str]): The whole file as a list of strings.
    start_index (int): The index in file_as_list where the token can be found.

    Returns:
    str:  The three part date as a string.
    """

    # Get the three parts of the date
    first_date_part = get_expected_token_value(token=initial_token, token_line=token_line,
                                               file_list=file_as_list)

    snipped_string = token_line.replace(initial_token, '')
    snipped_string = snipped_string.replace(first_date_part, '', 1)

    second_date_part = get_expected_next_value(start_line_index=start_index, file_as_list=file_as_list,
                                               search_string=snipped_string)

    snipped_string = snipped_string.replace(second_date_part, '', 1)

    third_date_part = get_next_value(start_line_index=start_index, file_as_list=file_as_list,
                                     search_string=snipped_string)

    full_date = f"{first_date_part} {second_date_part} {third_date_part}"
    return full_date


def get_expected_next_value(start_line_index: int, file_as_list: list[str], search_string: Optional[str] = None,
                            ignore_values: Optional[list[str]] = None,
                            replace_with: Union[str, GridArrayDefinition, None] = None,
                            custom_message: Optional[str] = None) -> str:
    """Gets the next non blank value in a list of lines.

    Args:
        start_line_index (int): line number to start reading file_as_list from
        file_as_list (list[str]): a list of strings containing each line of the file as a new entry
        search_string (str): string to search from within the first indexed line
        ignore_values (Optional[list[str]], optional): a list of values that should be ignored if found. \
            Defaults to None.
        replace_with (Union[str, VariableEntry, None], optional): a value to replace the existing value with. \
            Defaults to None.
        custom_message Optional[str]: A custom error message if no value is found

    Returns:
        str: Next non blank value from the list, if none found raises ValueError
    """
    value = get_next_value(start_line_index, file_as_list, search_string, ignore_values, replace_with)

    if value is None:
        if custom_message is None:
            raise ValueError(f"No value found in the line, line: {file_as_list[start_line_index]}")
        else:
            raise ValueError(f"{custom_message} {file_as_list[start_line_index]}")

    return value


def get_multiple_expected_sequential_values(list_of_strings: list[str], number_tokens: int,
                                            ignore_values: list[str]) -> list[str]:
    """Returns a sequential list of values as long as the number of tokens requested.

    Args:
        list_of_strings (list[str]): list of strings to represent the file with a new entry per line in the file.
        number_tokens (int): number of tokens to return values of
        ignore_values (list[str]): list of values to ignore if found

    Raises:
        ValueError: if too many tokens are requested compared to the file provided

    Returns:
        list[str]: list of strings comprised of the token values in order.
    """
    store_values: list[str] = []
    filter_list = list_of_strings.copy()
    for i in range(number_tokens):
        value = get_next_value(0, filter_list, filter_list[0], replace_with='',
                               ignore_values=ignore_values)
        if value is None:
            # if no valid value found, raise an error
            raise ValueError('Too many values requested from the list of strings passed,'
                             f' instead found: {len(store_values)} values, out of the requested {number_tokens}')
        store_values.append(value)

    return store_values


def get_nth_value(list_of_strings: list[str], value_number: int, ignore_values: list[str]) -> Optional[str]:
    """Returns the Nth value from a list of strings.

    Args:
        list_of_strings (list[str]): list of strings to represent the file with a new entry per line in the file.
        value_number (int): the index of the value that should be returned.One-based rather than zero-based.
        ignore_values (list[str]): list of values to ignore if found.

    Returns:
        Optional[str]: the value found at that location, if one is found
    """
    found_value: Optional[str] = None
    filter_list = list_of_strings.copy()

    if len(filter_list) == 0:
        return None

    for i in range(value_number):
        found_value = get_next_value(0, filter_list, filter_list[0], replace_with='',
                                     ignore_values=ignore_values)
        if found_value is None:
            return None

    return found_value



from enum import Enum


class PvtType(str, Enum):
    """Enum representing the different PVT types."""
    BLACKOIL = 'BLACKOIL'
    GASWATER = 'GASWATER'
    WATEROIL = 'WATEROIL'
    EOS = 'EOS'


class SeparatorType(str, Enum):
    """Enum representing the different separator types."""
    BLACKOIL = 'BLACKOIL'
    GASPLANT = 'GASPLANT'
    EOS = 'EOS'



"""Base class for handling any dynamic property simulator inputs, for use in inputs such as PVT, relperm, etc."""
from __future__ import annotations
from abc import ABC
from dataclasses import dataclass

import numpy as np
import pandas as pd
from typing import Optional
from ResSimpy.File import File
from ResSimpy.Units.AttributeMappings.BaseUnitMapping import BaseUnitMapping


@dataclass
class DynamicProperty(ABC):
    """The abstract base class for dynamic property simulator inputs, for use in inputs such as PVT, relperm, etc.

    Attributes:
        input_number (int): Method, table or input number, in order as entered in the simulation input deck.
    """

    input_number: int
    file: File
    properties: dict

    def __init__(self, input_number: int, file: File) -> None:
        """Initialises the DynamicProperty class.

        Args:
            input_number (int): Method, table or input number, in order as entered in the simulation input deck.
            file (File): The File that the dynamic property is read from.
        """
        self.input_number: int = input_number
        self.file: File = file

    @property
    def units(self) -> BaseUnitMapping:
        """Returns the attribute to unit map for the constraint."""
        raise NotImplementedError('Implement in the derived class.')

    def __repr__(self) -> str:
        """Pretty printing dynamic property data."""
        printable_str = f'\nFILE_PATH: {self.file.location}\n\n'
        printable_str += self.to_string()
        return printable_str

    def to_string(self) -> str:
        """Write dynamic property data to string."""
        raise NotImplementedError('Implement in the derived class.')

    def write_to_file(self, new_file_path: Optional[str] = None, overwrite_file: bool = False) -> None:
        """Write dynamic property data to file."""
        printable_str = self.to_string()
        new_file_contents = printable_str.splitlines(keepends=True)
        if overwrite_file is True and new_file_path is not None:
            raise ValueError('Please specify only one of either overwrite_existing or new_file_location.')

        if new_file_path is not None:
            new_file = File(file_content_as_list=new_file_contents, location=new_file_path, create_as_modified=True)
            new_file.write_to_file(new_file_path=new_file_path, overwrite_file=True)
            return
        elif overwrite_file is False:
            raise ValueError('Please specify either overwrite_file as True or provide new_file_location.')

        # Overwriting existing file contents
        if overwrite_file is True:
            self.file.file_content_as_list = new_file_contents
            self.file.write_to_file(overwrite_file=overwrite_file)

    @property
    def ranges(self) -> dict[str, tuple[float, float]]:
        """Returns a dictionary of the ranges of the dynamic properties."""
        ranges: dict[str, tuple[float, float]] = {}
        for prop, prop_value in self.properties.items():
            existing_range: tuple[float, float] = ranges.get(prop, (np.nan, np.nan))
            if isinstance(prop_value, pd.DataFrame):
                for col in prop_value.columns:
                    existing_range = ranges.get(col, (np.nan, np.nan))
                    ranges[col] = (np.nanmin((*existing_range, prop_value[col].min())),
                                   np.nanmax((*existing_range, prop_value[col].max())))
            elif isinstance(prop_value, str):
                # try to convert the string to a list of floats
                split_prop_value = prop_value.split()
                try:
                    split_prop_value_as_float = [float(val) for val in split_prop_value]
                except ValueError:
                    # if it can't be converted to a list of floats, then continue to the next property
                    continue
                if len(split_prop_value_as_float) == 0:
                    # skip if the list is empty
                    continue
                ranges[prop] = (np.nanmin((*existing_range, min(split_prop_value_as_float))),
                                np.nanmax((*existing_range, max(split_prop_value_as_float))))
            elif isinstance(prop_value, float):
                ranges[prop] = (prop_value, prop_value)
            else:
                continue
        return ranges



from dataclasses import dataclass
from abc import ABC
from typing import Literal, Sequence


@dataclass(kw_only=True)
class WellList:
    """Class for representing a single WellList or group for the model."""
    name: str
    wells: list[str]
    date: str


@dataclass(kw_only=True)
class WellLists(ABC):
    """Class for representing a set of WellList objects for the model."""
    __well_lists: Sequence[WellList]

    def __init__(self, welllists: Sequence[WellList]) -> None:
        """Initialises the WellLists class.

        Args:
            welllists (Sequence[WellList]): List of WellList objects.
        """
        self.__well_lists = welllists

    @property
    def _network_element_name(self) -> Literal['welllists']:
        return 'welllists'

    @property
    def welllists(self) -> Sequence[WellList]:
        """Returns all WellList instances."""
        return self.__well_lists

    @property
    def unique_names(self) -> list[str]:
        """Returns all WellList names."""
        return list({x.name for x in self.welllists})



def check_if_string_is_float(x: str) -> bool:
    """Function to check if a string can be successfully cast to a float.

    Args:
        x (str): String input to check

    Returns:
        bool: True if string can be successfully cast to a float, otherwise false.
    """
    try:
        float(x)
        return True
    except ValueError:
        return False



PREDEFINED_VARS = ["TIME", "ITIME", "NEWTON", "ALL1D", "ALL2D", "FIRST", "DAY", "MONTH", "YEAR", "SECOND", "MINUTE",
                   "HOUR", 'PRODWHCONS', "PRODBHCONS", "PRODWHNODES", "PRODBHNODES", "INJWHCONS", "INJBHCONS",
                   "INJWHNODES", "INJBHNODES", 'PRODWELLS', 'GASINJWELLS', 'WATERINJWELLS', 'CONSTRAINTS_CHANGED',
                   'NETWORK_CHANGED', 'NETWORK_TIME_INPUT', 'START_OF_RESTART']

CONDITIONAL_KEYWORDS = ['IF', 'ELSEIF', 'ELSE', 'THEN', 'EXIT', 'DO', 'CYCLE']

NEXUS_PROC_BASIC_FUNCS = ['ABS', 'ACCUMULATE', 'ANY', 'ALLOCATE', 'ALL', 'COUNT', 'MAX', 'MAXVAL', 'MIN', 'MINVAL',
                          'POWER', 'PRINTOUT', 'SET_VALUE', 'SIZE', 'SUM', 'EXP', 'LOG', 'LOG10', 'LOOKUP', 'OUTPUT',
                          'REDUCE', 'SPREADSHEET', 'SQRT']

NEXUS_PROC_PRINT_FUNCS = ['APPENDTOBUFFER', 'CLEARBUFFER', 'PRINTBUFFER', 'PRINTOUT', 'SETBUFFERPOSITION']

NEXUS_PROC_RETRIEVE_NETWORK_DATA_FUNCS = ['CONC', 'CONPARAM', 'CUM', 'GET_CONSTRAINT', 'GET_TARGET', 'GET_TGTCON',
                                          'GLR', 'GOR', 'IS_ACTIVE', 'IS_BACK_FLOWING', 'IS_FLOWING', 'IS_NOT_FLOWING',
                                          'IS_SHUTIN', 'OGR', 'P', 'PERF_LOOP', 'Q', 'QP', 'VALVE_SETTING', 'WAGPARAM',
                                          'WCUT', 'WGR', 'WOR', 'CALCULATE_AUTODRILL_BENEFIT', 'CONCAVG', 'CONINDEX',
                                          'CONINDICES', 'CONNECTEDDOWNSTREAM', 'CONNECTEDUPSTREAM', 'CPTFRAC',
                                          'CPTFRACP', 'CPTFRACS', 'CPTFRACSP', 'CPTRATE', 'CPTRATEP', 'CPTRATES',
                                          'CPPTRATESP', 'FIP', 'GASLIFTCONNECTION', 'GET_ACTIVE_PERFS', 'GET_COND',
                                          'GET_DEST', 'GET_DESTS', 'GET_DOWNSTREAM_MANIFOLD', 'GET_DOWNSTREAM_STATION',
                                          'GET_MAX_POSITION', 'GET_METHOD', 'GET_MIN_POSITION', 'GET_NODEIN',
                                          'GET_NODEOUT', 'GET_NUM_POSITIONS', 'GET_NUMBER_OF_PATTERNS', 'GET_OUTCONS',
                                          'GET_PATTERN_BALANCE', 'GET_PATTERN_INJECTION', 'GET_PATTERN_NUMBER',
                                          'GET_PATTERN_NUMBERS', 'GET_PATTERN_PRODUCTION', 'GET_PATTERN_RESERVOIR',
                                          'GET_PATTERN_RESERVOIRS', 'GET_PATTERN_WELLS', 'GET_POSITIONS', 'GET_PROP',
                                          'GET_QMULT', 'GET_STATION', 'GET_STATION_NODES', 'GET_STATION_PARENT',
                                          'GET_STATUS', 'GET_SYSTEM', 'GET_UPSTREAM_MANIFOLD', 'GET_UPSTREAM_STATION',
                                          'GET_VELOCITY', 'GET_WAGCYCLE', 'GET_WAGNCYCLE', 'GET_WAGSLUGSIZE',
                                          'GET_WELL_PATTERN', 'GET_WELL_PROP', 'GET_WELL_STATUS', 'GET_WELL_TYPE',
                                          'GET_WELLBH_METHOD', 'GET_WELLHEADCON', 'GET_WELLHEADNODE',
                                          'HASGASLIFTCONNECTION', 'HAS_POSITION', 'IS_ACTIVATED', 'LGR', 'NODEINDEX',
                                          'NODEINDICES', 'NUMBER_OF_WELLS_BROKEN', 'NUMBER_OF_WELLS_REPAIRED', 'PAVG',
                                          'PATTERN_ACTIVE', 'PATTERN_INDEX', 'PATTERN_INDICES', 'PDATAVG', 'PDATHCAVG',
                                          'PDATUM', 'PERF', 'PERF_PROP', 'PHCAVG', 'PP', 'RIGINDEX', 'RIGINDICES',
                                          'SALAVG', 'SAT', 'SATAVG', 'TAMB', 'TARGETINDEX', 'TARGETINDICES', 'TEMP',
                                          'VALVE_SETTING', 'WELLINDEX', 'WELLINDICES']


NEXUS_PROC_CHANGE_NETWORK_DATA_FUNS = ['ABANDON', 'ACTIVATE', 'ACTIVATE_NEXT', 'AUTODRILL', 'CONSTRAINT', 'DEACTIVATE',
                                       'DEACTIVATE_NEXT', 'DRILL', 'GLR_SHUT', 'NEXT_METHOD', 'QMULT', 'REDRILL',
                                       'RESET_RESERVOIR', 'SELECT_CON', 'SET_CONPARAM', 'SET_COND', 'SET_CONTEST',
                                       'SET_DEST', 'SET_METHOD', 'SET_PATTERN_BALANCE', 'SET_PDCORR', 'SET_PROP',
                                       'SET_RESERVOIR', 'SET_SERIES_METHOD', 'SET_STATION', 'SET_SYSTEM', 'SET_TAMB',
                                       'SET_TARGET', 'SET_TGTCON', 'SET_VALVE_SETTING', 'SET_WAGPARAM',
                                       'SET_WELL_METHOD', 'SET_WELL_PROP', 'SET_WELL_SERIES_METHOD', 'SHUTIN',
                                       'SOLVE_NETWORK', 'SOLVE_NETWORK_POTENTIAL', 'SOLVE_PARTIAL_POTENTIAL',
                                       'SOLVE_WELL_POTENTIAL', 'TEST_CON', 'TEST_WELL', 'UPDATE_NETWORK_CONFIG', 'WAG',
                                       'WCUTBACK', 'WELL_VALVE', 'WCUT_SHUT', 'WGR_SHUT', 'WOR_SHUT', 'WORKOVER',
                                       'WPLUG']


NEXUS_PROC_MANIP_LISTS_FUNCS = ["FILTER", "INTERSECT", "MASK", "UNION", "SORT"]

NEXUS_ALL_PROC_FUNCS = (NEXUS_PROC_BASIC_FUNCS + NEXUS_PROC_PRINT_FUNCS + NEXUS_PROC_MANIP_LISTS_FUNCS +
                        NEXUS_PROC_RETRIEVE_NETWORK_DATA_FUNCS + NEXUS_PROC_CHANGE_NETWORK_DATA_FUNS)



"""A class for collecting all NexusWell objects in a NexusSimulator. Handles adding and removing completions
as well as rescheduling wells.
"""
from __future__ import annotations
import warnings
from dataclasses import dataclass, field
from functools import cmp_to_key
from typing import Sequence, Optional, TYPE_CHECKING
from uuid import UUID

import pandas as pd

from ResSimpy.Enums.HowEnum import OperationEnum
from ResSimpy.Nexus.DataModels.NexusCompletion import NexusCompletion
from ResSimpy.Nexus.DataModels.NexusFile import NexusFile
from ResSimpy.Nexus.NexusEnums.DateFormatEnum import DateFormat
from ResSimpy.Nexus.NexusKeywords.wells_keywords import WELLS_KEYWORDS
from ResSimpy.Nexus.nexus_add_new_object_to_file import AddObjectOperations
from ResSimpy.Wells import Wells
from ResSimpy.Nexus.load_wells import load_wells
import ResSimpy.FileOperations.file_operations as fo
import ResSimpy.Nexus.nexus_file_operations as nfo
from ResSimpy.Utils.invert_nexus_map import attribute_name_to_nexus_keyword

if TYPE_CHECKING:
    from ResSimpy.Nexus.NexusSimulator import NexusSimulator
    from ResSimpy.Nexus.DataModels.NexusWell import NexusWell


@dataclass(kw_only=True)
class NexusWells(Wells):
    """A class for collecting all NexusWell objects in a NexusSimulator. Handles adding and removing completions
    as well as rescheduling wells. This is usually accessed through the model.wells property.

    Arguments:
        model (Simulator): NexusSimulator object that has the instance of wells on.
    """
    __model: NexusSimulator = field(compare=False, repr=False)
    __date_format: DateFormat = field(repr=False)
    _wells: list[NexusWell] = field(default_factory=list)

    def __init__(self, model: NexusSimulator) -> None:
        """Initialises the NexusWells class.

        Args:
            model (NexusSimulator): The model object that contains this NexusWells instance.
        """
        self.__model = model
        self.__add_object_operations = AddObjectOperations(NexusCompletion, self.table_header, self.table_footer, model)
        super().__init__()

    @property
    def model(self) -> NexusSimulator:
        """The model object that contains this NexusWells instance."""
        return self.__model

    @property
    def start_date(self) -> str:
        return self.__model.start_date

    @property
    def date_format(self) -> DateFormat:
        if not self._wells_loaded:
            self._load()
        return self.__date_format

    @property
    def table_header(self) -> str:
        return 'WELLSPEC'

    @property
    def table_footer(self) -> str:
        return ''

    def get_all(self) -> Sequence[NexusWell]:
        if not self._wells_loaded:
            self._load()
        return self._wells

    def get(self, well_name: str) -> Optional[NexusWell]:
        """Returns a specific well requested, or None if that well cannot be found."""
        if not self._wells_loaded:
            self._load()

        wells_to_return = filter(lambda x: x.well_name.upper() == well_name.upper(), self._wells)

        return next(wells_to_return, None)

    def get_df(self) -> pd.DataFrame:
        # loop through wells and completions to output a table
        if not self._wells_loaded:
            self._load()

        store_dictionaries = []
        for well in self._wells:
            for completion in well.completions:
                completion_props: dict[str, None | float | int | str] = {
                    'well_name': well.well_name,
                    'units': well.unit_system.name,
                }
                completion_props.update(completion.to_dict())
                store_dictionaries.append(completion_props)
        df_store = pd.DataFrame(store_dictionaries)
        df_store = df_store.dropna(axis=1, how='all')
        return df_store

    def _load(self) -> None:
        if self.__model.model_files.well_files is None:
            raise FileNotFoundError('No wells files found for current model.')
        for method_number, well_file in self.__model.model_files.well_files.items():
            if well_file.location is None:
                warnings.warn(f'Well file location has not been found for {well_file}')
                continue
            new_wells, date_format = load_wells(nexus_file=well_file, start_date=self.__model.start_date,
                                                default_units=self.__model.default_units, parent_wells_instance=self,
                                                model_date_format=self.__model.date_format)

            self._wells += new_wells
            self.__date_format = date_format
        self._wells_loaded = True

    def modify(self, well_name: str, completion_properties_list: list[dict[str, None | float | int | str]],
               how: OperationEnum = OperationEnum.ADD) -> None:
        """Modify the existing wells in memory using a dictionary of properties.

        Args:
        ----
            well_name (str): name of the well to modify
            completion_properties_list (list[InputDict]): a dictionary containing the properties to modify with the \
                attribute as keys and the values as the updated property value. If remove will remove perforation that \
                matches the values in the dictionary.
            how (OperationEnum): operation enum taking the values OperationEnum.ADD, OperationEnum.REMOVE. \
                Specifies how to modify the existing wells perforations.
            remove_all_that_match (bool): If True will remove all wells that partially match the completion_properties\
                provided. If False will remove perforation if only one matches, if several match throws a warning and \
                does not remove them.
            write_to_file (bool): If True writes directly to file. (Currently not in use)
        """
        well = self.get(well_name)
        if well is None:
            raise ValueError(f'No well named {well_name} found in simulator')
        for perf in completion_properties_list:
            if how == OperationEnum.ADD:
                try:
                    date = perf.get('date')
                except AttributeError:
                    raise AttributeError(
                        f'No date provided in perf: {perf}, please provide a date to add the perforation at.')
                if date is None:
                    raise AttributeError(
                        f'No date provided in perf: {perf}, please provide a date to add the perforation at.')
                self.add_completion(well_name=well_name, completion_properties=perf)
            elif how == OperationEnum.REMOVE:
                completions_to_remove = well.find_completions(perf)
                well._remove_completions_from_memory(completions_to_remove)
            elif how == OperationEnum.MODIFY:
                raise NotImplementedError('Modify in place not yet available. Please choose one of ADD/REMOVE')
            else:
                raise ValueError('Please select one of the valid OperationEnum values: e.g. OperationEnum.ADD')

    def add_completion(self, well_name: str, completion_properties: dict[str, None | float | int | str],
                       preserve_previous_completions: bool = True, comments: Optional[str] = None) -> None:
        """Adds a completion to an existing wellspec file.

        Args:
        ----
            well_name (str): well name to update
            completion_properties (dict[str, float | int | str]: properties of the completion you want to update.
            Must contain date of the completion to be added.
            preserve_previous_completions (bool): if true a new perforation added on a TIME card without a \
            wellspec card for that well will preserve the previous completions from the closest TIME card in addition \
            to the new completion
        """
        basic_dict: dict[str, float | int | str | None] = {'name': well_name}
        _, completion_date = self.__add_object_operations.check_name_date(basic_dict | completion_properties)
        well = self.get(well_name)
        if well is None:
            # TODO could make this not raise an error and instead initialize a NexusWell and add it to NexusWells
            raise ValueError(
                f"No well found named: {well_name}. Cannot add completion to a well that doesn't exist")
        well_id = well.completions[0].id

        # add completion in memory
        new_completion = well._add_completion_to_memory(date=completion_date,
                                                        completion_properties=completion_properties,
                                                        date_format=self.date_format)

        if self.__model.model_files.well_files is None:
            raise FileNotFoundError('No well file found, cannot modify ')

        wellspec_file = self.__add_object_operations.find_which_file_from_id(obj_id=well_id,
                                                                             file_type_to_search='well_files')

        # initialise some storage variables
        nexus_mapping = NexusCompletion.get_keyword_mapping()
        new_completion_time_index = -1
        header_index = -1
        headers: list[str] = []
        headers_original: list[str] = []
        additional_headers: list[str] = []
        file_content = wellspec_file.get_flat_list_str_file
        date_found = False
        new_completion_index = len(file_content)
        new_completion_string: list[str] = []
        last_valid_line_index: int = -1
        writing_new_wellspec_table = False
        date_comp = 0
        # if no time cards present in the file just find the name of the well instead
        if not fo.value_in_file('TIME', file_content):
            new_completion_time_index = 0

        for index, line in enumerate(file_content):
            if header_index == -1 and fo.check_token('TIME', line):
                wellspec_date = fo.get_expected_token_value('TIME', line, [line])
                date_comp = self.__model._sim_controls.compare_dates(wellspec_date, completion_date)
                if date_comp == 0:
                    # if we've found the date we're looking for start looking for a wellspec and name card
                    new_completion_time_index = index
                    date_found = True
                    continue
                elif date_comp > 0 and header_index == -1:
                    # if no date is found that is equal and we've found a date that is greater than the specified date
                    # start to compile a wellspec table from scratch and add in the time cards
                    new_completion_index = index
                    header_index = index - 1
                    headers, new_completion_index, new_completion_string, found_completion_at_previous_date = \
                        self.__write_out_existing_wellspec(
                            completion_date, completion_properties, date_found, index,
                            new_completion_index, preserve_previous_completions, well, well_name)
                    writing_new_wellspec_table = True
                    if not found_completion_at_previous_date:
                        break
                else:
                    continue

            if fo.check_token('WELLSPEC', line) and fo.get_token_value('WELLSPEC', line, [line]) == well_name \
                    and new_completion_time_index != -1:
                # get the header of the wellspec table
                header_index, headers, headers_original = self.__add_object_operations.get_and_write_new_header(
                    additional_headers, completion_properties, file_content, index, nexus_mapping, wellspec_file
                )
                continue

            if header_index != -1 and nfo.nexus_token_found(line, WELLS_KEYWORDS) and index > header_index:
                # if we hit the end of the wellspec table for the given well set the index for the new completion
                if last_valid_line_index != -1:
                    new_completion_index = last_valid_line_index + 1
                break

            elif header_index != -1 and index > header_index:
                # check for valid rows + fill extra columns with NA
                line_valid_index = self.__add_object_operations.fill_in_nas(additional_headers, headers_original, index,
                                                                            line, wellspec_file, file_content)
                # set the line to insert the new completion at to be the one after the last valid line
                last_valid_line_index = line_valid_index if line_valid_index > 0 else last_valid_line_index
        # If we haven't found a TIME card after the for loop then we haven't got a valid date so add it at the end
        if date_comp < 0:
            new_completion_index = len(file_content)
            headers, new_completion_index, new_completion_string, found_completion_at_previous_date = \
                self.__write_out_existing_wellspec(
                    completion_date, completion_properties, date_found, new_completion_index,
                    new_completion_index, preserve_previous_completions, well, well_name)
            writing_new_wellspec_table = True

        # construct the new completion and ensure the order of the values is in the same order as the headers
        new_completion_string += [new_completion.to_table_line(headers)]
        new_completion_additional_lines = len(new_completion_string)
        if writing_new_wellspec_table:
            new_completion_string += ['\n']
        # write out to the file_content_as_list
        new_completion_object_ids = {new_completion.id: [new_completion_index + new_completion_additional_lines - 1]}
        wellspec_file.add_to_file_as_list(additional_content=new_completion_string, index=new_completion_index,
                                          additional_objects=new_completion_object_ids, comments=comments)

    def __write_out_existing_wellspec(self, completion_date: str,
                                      completion_properties: dict[str, None | float | int | str],
                                      date_found: bool, index: int, new_completion_index: int,
                                      preserve_previous_completions: bool, well: NexusWell, well_name: str) -> \
            tuple[list[str], int, list[str], bool]:
        """Writes out the existing wellspec for a well at a new time stamp."""
        nexus_mapping = NexusCompletion.get_keyword_mapping()
        completion_table_as_list = ['\n']
        if not date_found:
            completion_table_as_list += ['TIME ' + completion_date + '\n']
        completion_table_as_list += ['WELLSPEC ' + well_name + '\n']
        headers = [k for k, v in nexus_mapping.items() if v[0] in completion_properties]
        if preserve_previous_completions:
            # get all the dates for that well
            date_list = well.dates_of_completions
            previous_dates = [x for x in date_list if
                              self.__model._sim_controls.compare_dates(x, completion_date) < 0]
            if len(previous_dates) == 0:
                # if no dates that are smaller than the completion date then only add the perforation
                # at the current index with a new wellspec card.
                warnings.warn(f'No previous completions found for {well_name} at date: {completion_date}')
                new_completion_index = index
                write_out_headers = [' '.join(headers) + '\n']
                completion_table_as_list += write_out_headers
                return headers, new_completion_index, completion_table_as_list, False

            # get the most recent date that is earlier than the new completion date
            previous_dates = sorted(previous_dates, key=cmp_to_key(self.__model._sim_controls.compare_dates))
            last_date = str(previous_dates[-1])
            completion_to_find: dict[str, None | float | int | str] = {'date': last_date}
            # find all completions at this date
            previous_completion_list = well.find_completions(completion_to_find)
            if len(previous_completion_list) > 0:
                prev_completion_properties = {k: v for k, v in previous_completion_list[0].to_dict().items() if
                                              v is not None}
                for key in prev_completion_properties:
                    if key == 'date':
                        continue
                    header_key = attribute_name_to_nexus_keyword(nexus_mapping, key)
                    if header_key not in headers:
                        headers.append(header_key)

            write_out_headers = [' '.join(headers) + '\n']
            completion_table_as_list += write_out_headers
            # run through the existing completions to duplicate the completion at the new time
            for completion in previous_completion_list:
                completion_table_as_list += [completion.to_table_line(headers)]
        else:
            write_out_headers = [' '.join(headers) + '\n']
            completion_table_as_list += write_out_headers
        return headers, new_completion_index, completion_table_as_list, True

    def remove_completion(self, well_name: str,
                          completion_properties: Optional[dict[str, None | float | int | str]] = None,
                          completion_id: Optional[UUID] = None) -> None:

        well = self.get(well_name)
        if well is None:
            raise ValueError(f'No well found with name: {well_name}')

        if completion_properties is None and completion_id is None:
            raise ValueError('Must provide one of completion_properties dictionary or completion_id.')

        # check for a date:
        if completion_properties is not None:
            completion_date = completion_properties.get('date', 'NO_DATE_PROVIDED')
            if completion_date == 'NO_DATE_PROVIDED':
                raise AttributeError('Completion requires a date. '
                                     'Please provide a date in the completion_properties_list dictionary.')
            if completion_id is None:
                completion_id = well.find_completion(completion_properties).id
        if completion_id is None:
            raise ValueError('No completion found for completion_properties')
        # find which wellspec file we should edit
        wellspec_file = self.__add_object_operations.find_which_file_from_id(obj_id=completion_id,
                                                                             file_type_to_search='well_files')

        # remove from the well object/wells class
        completion_date = well.get_completion_by_id(completion_id).date
        well._remove_completion_from_memory(completion_to_remove=completion_id)

        # drop it from the wellspec file or include file if stored in include file
        if wellspec_file.object_locations is None:
            raise ValueError(f'No object locations specified, cannot find completion id: {completion_id}')
        completion_indices = wellspec_file.object_locations[completion_id]
        if len(completion_indices) > 0:
            for comp_index in completion_indices:
                wellspec_file.remove_from_file_as_list(comp_index, [completion_id])

        # check that we have completions left:
        find_completions_dict: dict[str, None | float | int | str] = {'date': completion_date}
        remaining_completions = well.find_completions(find_completions_dict)
        if len(remaining_completions) == 0:
            # if there are no more completions remaining for that time stamp then remove the wellspec header!
            self.__remove_wellspec_header(str(completion_date), well_name, wellspec_file)

    def __remove_wellspec_header(self, completion_date: str, well_name: str, wellspec_file: NexusFile) -> None:
        """Removes the wellspec and header if the wellspec table is empty\
        must first check for whether the well has any remaining completions in the wellspec table.
        """
        nexus_mapping = NexusCompletion.get_keyword_mapping()
        completion_date_found = False
        file_content = wellspec_file.get_flat_list_str_file
        wellspec_index = -1
        header_index = -1
        if not fo.value_in_file('TIME', file_content):
            # if we have no completion date in the file then we have effectively found the right TIME
            completion_date_found = True
        for index, line in enumerate(file_content):
            if fo.check_token('TIME', line) and fo.get_expected_token_value('TIME', line, [line]) == \
                    completion_date:
                completion_date_found = True
            if completion_date_found and fo.check_token('WELLSPEC', line) and \
                    fo.get_token_value('WELLSPEC', line, [line]) == well_name:
                # get the index in the list as string
                wellspec_index = index
                keyword_map = {x: y[0] for x, y in nexus_mapping.items()}
                wellspec_table = file_content[wellspec_index::]
                header_index, _ = nfo.get_table_header(file_as_list=wellspec_table, header_values=keyword_map)
                header_index += wellspec_index
                break
        wellspec_file.remove_from_file_as_list(header_index)
        wellspec_file.remove_from_file_as_list(wellspec_index)

    def modify_completion(self, well_name: str, properties_to_modify: dict[str, None | float | int | str],
                          completion_to_change: Optional[dict[str, None | float | int | str]] = None,
                          completion_id: Optional[UUID] = None,
                          comments: Optional[str] = None) -> None:
        """Modify an existing matching completion, preserves attributes and modifies only additional properties
        found within the provided properties to modify dictionary.

        Args:
            well_name (str): Name of the well with the completion to be modified.
            properties_to_modify (dict[str, None | float | int | str]): attributes to change to.
            completion_to_change (Optional[dict[str, None | float | int | str]]): properties of the existing completion.
            User must provide enough to uniquely identify the completion.
            completion_id (Optional[UUID]): If provided will match against a known UUID for the completion.
        """
        well = self.get(well_name)
        if well is None:
            raise ValueError(f'No well found with name: {well_name}')

        if completion_to_change is not None:
            completion = well.find_completion(completion_to_change)
            completion_id = completion.id
        elif completion_id is not None:
            completion = well.get_completion_by_id(completion_id)
        else:
            raise ValueError('Must provide one of completion_to_change dictionary or completion_id')

        # start with the existing properties
        update_completion_properties: dict[str, None | float | int | str] = \
            {k: v for k, v in completion.to_dict().items() if v is not None}

        update_completion_properties.update(properties_to_modify)

        self.remove_completion(well_name, completion_id=completion_id)
        self.add_completion(well_name, update_completion_properties, preserve_previous_completions=True,
                            comments=comments)



from __future__ import annotations
from typing import Optional, TYPE_CHECKING

import pandas as pd
from ResSimpy.Grid import GridArrayDefinition
from ResSimpy.Nexus.NexusKeywords.structured_grid_keywords import GRID_ARRAY_KEYWORDS, STRUCTURED_GRID_KEYWORDS
from ResSimpy.Nexus.NexusKeywords.structured_grid_keywords import GRID_ARRAY_FORMAT_KEYWORDS
import ResSimpy.Nexus.nexus_file_operations as nfo
import ResSimpy.FileOperations.file_operations as fo
from ResSimpy.Utils.general_utilities import check_if_string_is_float

if TYPE_CHECKING:
    from ResSimpy.Nexus.NexusSimulator import NexusSimulator


class StructuredGridOperations:
    def __init__(self, model: NexusSimulator) -> None:
        """Initialises the StructuredGridOperations class.

        Args:
            model (NexusSimulator): Parent model object which the StructuredGridOperations class is associated with.
        """
        self.__model: NexusSimulator = model

    @staticmethod
    def load_token_value_if_present(token: str, modifier: str,
                                    token_property: GridArrayDefinition | dict[str, GridArrayDefinition],
                                    line: str, file_as_list: list[str], line_indx: int,
                                    ignore_values: Optional[list[str]] = None) -> None:
        """Gets a token's value if there is one and loads it into the token_property.

        Args:
        ----
            token (str): the token being searched for. e.g. 'PERMX'
            modifier (str): any modifiers applied to the token e.g. 'MULT'
            token_property (GridArrayDefinition | dict[str, GridArrayDefinition]):
                GridArrayDefinition or dict of GridArrayDefinition object to store the modifier and value pair into
            line (str): line to search for the token in
            line_indx (int): index of line in file_as_list
            file_as_list (list[str]): a list of strings containing each line of the file as a new entry
            ignore_values (Optional[list[str]]): values to be ignored. Defaults to None.

        Raises:
        ------
            ValueError: raises an error if no numerical value can be found after the supplied token modifier pair
        Returns:
            None: stores the value into token_property supplied instead
        """

        if ignore_values is None:
            ignore_values = []
        region_name: str = ''
        modifier_found = False
        token_found = nfo.check_token(token, line)
        found_modifier = line.strip().split()[-1]
        token_modifier: str = ''

        if token_found:
            token_next_value = fo.get_token_value(token, line, file_as_list, ignore_values=ignore_values)
        else:
            return
        if token_next_value == found_modifier:
            modifier_found = token_next_value == modifier
            token_modifier = f"{token} {modifier}"
        # If IREGION and the region_group is named
        elif token == 'IREGION' and found_modifier in GRID_ARRAY_FORMAT_KEYWORDS and len(line.strip().split()) == 3:
            region_name = line.strip().split()[1]
            modifier_found = found_modifier == modifier
            token_modifier = f"{token} {region_name} {modifier}"
        if token == 'IREGION' and isinstance(token_property, dict):
            if region_name == '':
                region_name = f"IREG{len(token_property.keys())+1}"

        if token_found and modifier_found:
            # If we are loading a multiple, load the two relevant values, otherwise just the next value
            # once we have the region_name, modifier_found, and token_modifier we can break it up in three pieces
            # piece one: make the grid definition array
            # piece two: find out if we have any MOD cards to deal with
            # piece three: make the mod table and add it to the grid array
            StructuredGridOperations.__make_grid_def(file_as_list, ignore_values, line, line_indx, modifier,
                                                     region_name, token_modifier, token_property)

            mod_start_end = StructuredGridOperations.__extract_mod_positions(line_indx, file_as_list)
            StructuredGridOperations.__make_mod_table(mod_start_end, file_as_list,
                                                      line, token_property, region_name, token_modifier)

    @staticmethod
    def __make_grid_def(file_as_list: list[str], ignore_values: list[str], line: str, line_indx: int, modifier: str,
                        region_name: str, token_modifier: str,
                        token_property: GridArrayDefinition | dict[str, GridArrayDefinition]) -> None:
        """A function that begins the process of populating the grid away."""

        if modifier == 'MULT':
            numerical_value = nfo.get_expected_token_value(modifier, line, file_as_list, ignore_values=ignore_values)
            if numerical_value is None:
                raise ValueError(
                    f'No numerical value found after {token_modifier} keyword in line: {line}')
            value_to_multiply = fo.get_token_value(modifier, line, file_as_list,
                                                   ignore_values=[numerical_value, *ignore_values])
            if numerical_value is not None and value_to_multiply is not None:
                if not isinstance(token_property, dict):
                    token_property.modifier = 'MULT'
                    token_property.value = f"{numerical_value} {value_to_multiply}"
                elif region_name != '':  # IREGION
                    token_property[region_name] = GridArrayDefinition()
                    token_property[region_name].modifier = 'MULT'
                    token_property[region_name].value = f"{numerical_value} {value_to_multiply}"
        else:
            value = fo.get_token_value(modifier, line, file_as_list, ignore_values=ignore_values)
            if value is None:
                # Could be 'cut short' by us excluding the rest of a file.
                if not isinstance(token_property, dict):
                    token_property.value = None
                    token_property.modifier = modifier
                elif region_name != '':  # IREGION
                    token_property[region_name] = GridArrayDefinition()
                    token_property[region_name].modifier = modifier
                    token_property[region_name].value = None
            else:
                # Check if VALUE modifier is followed by a numeric string or an INCLUDE file
                line_to_check = file_as_list[line_indx + 1]
                # Check next couple of lines in case next line contains one of ignore_values
                for i in range(line_indx+1, min(line_indx+3, len(file_as_list))):
                    if value in file_as_list[i]:
                        line_to_check = file_as_list[i]
                if 'INCLUDE' in line_to_check:
                    if not isinstance(token_property, dict):
                        token_property.value = value
                        token_property.modifier = modifier
                    elif region_name != '':  # IREGION
                        token_property[region_name] = GridArrayDefinition()
                        token_property[region_name].modifier = modifier
                        token_property[region_name].value = value
                elif check_if_string_is_float(value[0]) or value[0] == '.':
                    start_indx = line_indx + 1
                    end_indx = len(file_as_list)
                    found_end_value = False
                    for i in range(start_indx, len(file_as_list)):
                        for keyword in STRUCTURED_GRID_KEYWORDS + GRID_ARRAY_KEYWORDS:
                            if nfo.check_token(keyword, file_as_list[i]):
                                end_indx = i
                                found_end_value = True
                                break
                        if found_end_value:
                            break
                    if not isinstance(token_property, dict):
                        token_property.modifier = modifier
                        token_property.value = '\n'.join([line.strip()
                                                          for line in file_as_list[start_indx:end_indx]]).strip()
                    elif region_name != '':  # IREGION
                        token_property[region_name] = GridArrayDefinition()
                        token_property[region_name].modifier = modifier
                        token_property[region_name].value = '\n'.join([line.strip() for line in
                                                                       file_as_list[start_indx:end_indx]]).strip()
                else:  # The grid array keyword is likely inside an include file, presented on previous line
                    if 'INCLUDE' in file_as_list[line_indx - 1]:
                        if not isinstance(token_property, dict):
                            token_property.modifier = modifier
                            token_property.value = file_as_list[line_indx - 1].split('INCLUDE')[1].strip()
                            token_property.keyword_in_include_file = True
                        elif region_name != '':  # IREGION
                            token_property[region_name] = GridArrayDefinition()
                            token_property[region_name].modifier = modifier
                            token_property[region_name].value \
                                = file_as_list[line_indx - 1].split('INCLUDE')[1].strip()
                            token_property[region_name].keyword_in_include_file = True
                    else:
                        raise ValueError(
                            f'No suitable value found after {token_modifier} keyword in line: {line}')

    @staticmethod
    def replace_value(file_as_list: list[str], old_property: GridArrayDefinition, new_property: GridArrayDefinition,
                      token_name: str) -> None:
        """Replace the value and token + modifier with the new values.

        Args:
        ----
            file_as_list (list[str]): a list of strings containing each line of the file as a new entry
            old_property (VariableEntry): property found in the original file to be replaced
            new_property (VariableEntry): new property to replace the old property with
            token_name (str): name of the token being replaced
        Returns:
            None: modifies the file_as_list with the new property
        """

        for line in file_as_list:
            old_token_modifier = f"{token_name} {old_property.modifier}"
            new_token_modifier = f"{token_name} {new_property.modifier}"
            ignore_values = ['INCLUDE'] if old_property.modifier == 'VALUE' else []
            if nfo.check_token(old_token_modifier, line):
                # If we are replacing a mult, replace the first value with a blank
                if old_property.modifier == 'MULT':
                    dummy_value = GridArrayDefinition('MULT', '')
                    fo.get_token_value(old_token_modifier, line, file_as_list, ignore_values=ignore_values,
                                       replace_with=dummy_value)

                fo.get_token_value(old_token_modifier, line, file_as_list, ignore_values=ignore_values,
                                   replace_with=new_property)

                new_line = line.replace(old_token_modifier, new_token_modifier, 1)
                line_index = file_as_list.index(line)
                file_as_list[line_index] = new_line

    @staticmethod
    def append_include_to_grid_file(include_file_location: str, structured_grid_file_path: str) -> None:
        # TODO: change append to be an optional parameter
        """Appends an include file to the end of a grid for adding LGRs.

        Args:
        ----
            include_file_location (str): path to a file to include in the grid
            structured_grid_file_path (str): file path to the structured grid.

        Raises:
        ------
            ValueError: if no structured grid file path is specified in the class instance
        """
        # Get the existing file as a list
        if structured_grid_file_path is None:
            raise ValueError("No file path given or found for structured grid file path. \
                Please update structured grid file path")
        file = nfo.load_file_as_list(structured_grid_file_path)

        file.append(f"\nINCLUDE {include_file_location}\n")
        file.append("TOLPV LGR1 0\n")

        # Save the new file contents
        new_file_str = "".join(file)
        with open(structured_grid_file_path, "w") as text_file:
            text_file.write(new_file_str)

    @staticmethod
    def get_grid_file_as_3d_list(path: str) -> Optional[list]:
        """Converts a grid file to a 3D list.

        Args:
        ----
            path (str): path to a grid file

        Returns:
        -------
            Optional[list[str]]: Returns None if no file is found, returns the grid as a 3d array otherwise
        """
        try:
            with open(path) as f:
                grid_file_list = list(f)
        except FileNotFoundError:
            return None

        sub_lists = []

        new_list_str = ""
        for sub_list in grid_file_list[4:]:
            if sub_list == '\n':
                new_list_split = [x.split("\t") for x in new_list_str.split("\n")]
                new_list_split_cleaned = []
                for x_list in new_list_split:
                    float_list_split = [float(x) for x in x_list if x != ""]
                    new_list_split_cleaned.append(float_list_split)
                sub_lists.append(new_list_split_cleaned)
                new_list_str = ""
            else:
                new_list_str = new_list_str + sub_list
        return sub_lists

    def view_command(self, field: str, previous_lines: int = 3, following_lines: int = 3) -> Optional[str]:
        """Displays how the property is declared in the structured grid file.

        Args:
        ----
            field (str): property as written in the structured grid (e.g. KX)
            previous_lines (int, optional): how many lines to look back from the field searched for. Defaults to 3.
            following_lines (int, optional): how many lines to look forward from the field searched for. Defaults to 3.

        Raises:
        ------
            ValueError: if no structured grid file path is specified in the class instance

        Returns:
        -------
            Optional[str]: the string associated with the supplied property from within the structured grid. \
                If the field is not found in the structured grid returns None.
        """
        structured_grid_dict = self.__model.get_structured_grid_dict()
        structured_grid_path = self.__model.structured_grid_path
        command_token = f"{field.upper()} {structured_grid_dict[field.lower()].modifier}"
        if structured_grid_path is None:
            raise ValueError("No path found for structured grid file path. \
                Please provide a path to the structured grid")
        file_as_list = nfo.load_file_as_list(structured_grid_path)

        for line in file_as_list:
            if nfo.check_token(command_token, line):
                start_index = file_as_list.index(line) - previous_lines \
                    if file_as_list.index(line) - previous_lines > 0 else 0
                end_index = file_as_list.index(line) + following_lines \
                    if file_as_list.index(line) + following_lines < len(file_as_list) else len(file_as_list) - 1

                new_array = file_as_list[start_index: end_index]
                new_array = [x.strip("'") for x in new_array]
                value = "".join(new_array)
                return value
        return None

    @staticmethod
    def __extract_mod_positions(line_indx: int, file_as_list: list[str]) -> dict[str, list[list[int]]]:
        """Finds the positions where MOD exists."""

        mod_start_end: dict[str, list[list[int]]] = {}
        break_flag = False
        for i in range(line_indx + 1, len(file_as_list)):
            if nfo.check_token('MODX', file_as_list[i]):
                mod_start_end['MODX'] = [[i + 1, i + 2]]
            if nfo.check_token('MODY', file_as_list[i]):
                mod_start_end['MODY'] = [[i + 1, i + 2]]
            if nfo.check_token('MODZ', file_as_list[i]):
                mod_start_end['MODZ'] = [[i + 1, i + 2]]
            if nfo.check_token('MOD', file_as_list[i]):
                if 'MOD' in mod_start_end.keys():  # Already found a prior mod for this token, append
                    mod_start_end['MOD'].append([i + 1, len(file_as_list)])
                else:
                    mod_start_end['MOD'] = [[i + 1, len(file_as_list)]]
                found_end_of_mod_table = False
                for j in range(i + 1, len(file_as_list)):
                    for keyword in STRUCTURED_GRID_KEYWORDS + GRID_ARRAY_KEYWORDS:
                        if nfo.check_token(keyword, file_as_list[j]):
                            mod_start_end['MOD'][-1][1] = j
                            found_end_of_mod_table = True
                            break
                    if found_end_of_mod_table:
                        break
            for keyword in GRID_ARRAY_KEYWORDS:
                if nfo.check_token(keyword, file_as_list[i]):
                    break_flag = True
                    break
            if break_flag:
                break

        return mod_start_end

    @staticmethod
    def __make_mod_table(mod_start_end: dict[str, list[list[int]]], file_as_list: list[str], line: str,
                         token_property: GridArrayDefinition | dict[str, GridArrayDefinition], region_name: str,
                         token_modifier: str) -> None:
        """A function that creates the mod table."""

        for key in mod_start_end.keys():
            for i in range(len(mod_start_end[key])):
                if mod_start_end[key][i][1] > mod_start_end[key][i][0]:
                    mod_table = nfo.read_table_to_df(
                        file_as_list[mod_start_end[key][i][0]:mod_start_end[key][i][1]], noheader=True)
                    if len(mod_table.columns) == 7:
                        mod_table.columns = ['i1', 'i2', 'j1', 'j2', 'k1', 'k2', '#v']
                    elif len(mod_table.columns) == 8:
                        mod_table[8] = mod_table[6].astype(str) + mod_table[7].astype(str)
                        mod_table = mod_table.drop([6, 7], axis=1)
                        mod_table.columns = ['i1', 'i2', 'j1', 'j2', 'k1', 'k2', '#v']
                    else:
                        raise ValueError(
                            f'Unsuitable mod card for {token_modifier} keyword in line: {line}')
                    if not isinstance(token_property, dict):
                        if token_property.mods is not None:
                            if key in token_property.mods.keys():
                                orig_mod_tab = token_property.mods[key].copy()
                                token_property.mods[key] = \
                                    pd.concat([orig_mod_tab, mod_table]).reset_index(drop=True)
                            else:
                                token_property.mods[key] = mod_table
                        else:
                            token_property.mods = {key: mod_table}
                    else:  # IREGION
                        if token_property[region_name].mods is not None:
                            tmp_dict = token_property[region_name].mods
                            if isinstance(tmp_dict, dict):
                                if key in tmp_dict.keys():
                                    orig_mod_tab = tmp_dict[key].copy()
                                    tmp_dict[key] = \
                                        pd.concat([orig_mod_tab, mod_table]).reset_index(drop=True)
                        else:
                            token_property[region_name].mods = {key: mod_table}



"""Nexus grid file class for loading in a structured grid file and extracting the grid properties."""
from __future__ import annotations

import copy
import pandas as pd
from dataclasses import dataclass
from typing import Optional, TYPE_CHECKING, Any
import warnings

from ResSimpy.File import File
from ResSimpy.Grid import Grid, GridArrayDefinition
from ResSimpy.Nexus.DataModels.NexusFile import NexusFile
from ResSimpy.Nexus.DataModels.StructuredGrid.NexusGridArrayFunction import NexusGridArrayFunction
from ResSimpy.Nexus.NexusKeywords.structured_grid_keywords import GRID_ARRAY_FORMAT_KEYWORDS
from ResSimpy.Nexus.structured_grid_operations import StructuredGridOperations
import ResSimpy.Nexus.nexus_file_operations as nfo
import ResSimpy.Nexus.array_function_operations as afo

from resqpy.olio.read_nexus_fault import load_nexus_fault_mult_table_from_list

if TYPE_CHECKING:
    from ResSimpy.Nexus.NexusSimulator import NexusSimulator


@dataclass
class PropertyToLoad:
    token: str
    modifiers: list[str]
    property: GridArrayDefinition | dict[str, GridArrayDefinition]


@dataclass(kw_only=True)
class NexusGrid(Grid):
    __array_functions_list: Optional[list[list[str]]] = None
    __array_functions_df: Optional[pd.DataFrame] = None
    __array_functions_loaded: bool = False
    __grid_file_contents: Optional[list[str]] = None
    __grid_file_nested: Optional[list[str]] = None
    __faults_df: Optional[pd.DataFrame] = None
    __grid_faults_loaded: bool = False
    __grid_nexus_file: Optional[NexusFile] = None
    __corp: GridArrayDefinition
    __dx: GridArrayDefinition
    __dy: GridArrayDefinition
    __dz: GridArrayDefinition
    __depth: GridArrayDefinition
    __mdepth: GridArrayDefinition
    __dznet: GridArrayDefinition
    __compr: GridArrayDefinition
    __icoars: GridArrayDefinition
    __ialphaf: GridArrayDefinition
    __ipolymer: GridArrayDefinition
    __iadsorption: GridArrayDefinition
    __itracer: GridArrayDefinition
    __igrid: GridArrayDefinition
    __isector: GridArrayDefinition
    __swl: GridArrayDefinition
    __swr: GridArrayDefinition
    __swu: GridArrayDefinition
    __sgl: GridArrayDefinition
    __sgr: GridArrayDefinition
    __sgu: GridArrayDefinition
    __swro: GridArrayDefinition
    __swro_ls: GridArrayDefinition
    __sgro: GridArrayDefinition
    __sgrw: GridArrayDefinition
    __krw_swro: GridArrayDefinition
    __krws_ls: GridArrayDefinition
    __krw_swu: GridArrayDefinition
    __krg_sgro: GridArrayDefinition
    __krg_sgu: GridArrayDefinition
    __krg_sgrw: GridArrayDefinition
    __kro_swl: GridArrayDefinition
    __kro_swr: GridArrayDefinition
    __kro_sgl: GridArrayDefinition
    __kro_sgr: GridArrayDefinition
    __krw_sgl: GridArrayDefinition
    __krw_sgr: GridArrayDefinition
    __sgtr: GridArrayDefinition
    __sotr: GridArrayDefinition
    __swlpc: GridArrayDefinition
    __sglpc: GridArrayDefinition
    __pcw_swl: GridArrayDefinition
    __pcg_sgu: GridArrayDefinition
    __chloride: GridArrayDefinition
    __calcium: GridArrayDefinition
    __salinity: GridArrayDefinition
    __api: GridArrayDefinition
    __tmx: GridArrayDefinition
    __tmy: GridArrayDefinition
    __tmz: GridArrayDefinition
    __multbv: GridArrayDefinition
    __pv: GridArrayDefinition
    __iequil: GridArrayDefinition
    __ipvt: GridArrayDefinition
    __iwater: GridArrayDefinition
    __irelpm: GridArrayDefinition
    __irock: GridArrayDefinition
    __itran: GridArrayDefinition
    __iregion: dict[str, GridArrayDefinition]
    __pvmult: GridArrayDefinition
    __livecell: GridArrayDefinition
    __worka1: GridArrayDefinition
    __worka2: GridArrayDefinition
    __worka3: GridArrayDefinition
    __worka4: GridArrayDefinition
    __worka5: GridArrayDefinition
    __worka6: GridArrayDefinition
    __worka7: GridArrayDefinition
    __worka8: GridArrayDefinition
    __worka9: GridArrayDefinition

    def __init__(self, grid_nexus_file: Optional[NexusFile] = None, assume_loaded: bool = False) -> None:
        """Initialises the NexusGrid class.

        Args:
            grid_nexus_file (Optional[NexusFile]): the NexusFile representation of a structured grid file for \
                reading and interpreting the grid properties from.
        """
        super().__init__(assume_loaded=assume_loaded)
        self.__array_functions_list: Optional[list[list[str]]] = None
        self.__array_functions_df: Optional[pd.DataFrame] = None
        self.__array_functions_loaded: bool = False
        self.__grid_file_contents: Optional[list[str]] = None if grid_nexus_file is None else \
            grid_nexus_file.get_flat_list_str_file_including_includes
        self.__grid_file_nested: Optional[list[str]] = None if grid_nexus_file is None else \
            grid_nexus_file.file_content_as_list
        self.__faults_df: Optional[pd.DataFrame] = None
        self.__grid_nexus_file: Optional[NexusFile] = grid_nexus_file
        self.__grid_array_functions: Optional[list[NexusGridArrayFunction]] = None
        self.__corp: GridArrayDefinition = GridArrayDefinition()
        self.__iequil: GridArrayDefinition = GridArrayDefinition()
        self.__ipvt: GridArrayDefinition = GridArrayDefinition()
        self.__iwater: GridArrayDefinition = GridArrayDefinition()
        self.__irelpm: GridArrayDefinition = GridArrayDefinition()
        self.__irock: GridArrayDefinition = GridArrayDefinition()
        self.__itran: GridArrayDefinition = GridArrayDefinition()
        self.__iregion: dict[str, GridArrayDefinition] = {}
        self.__pvmult: GridArrayDefinition = GridArrayDefinition()
        self.__livecell: GridArrayDefinition = GridArrayDefinition()
        self.__worka1: GridArrayDefinition = GridArrayDefinition()
        self.__worka2: GridArrayDefinition = GridArrayDefinition()
        self.__worka3: GridArrayDefinition = GridArrayDefinition()
        self.__worka4: GridArrayDefinition = GridArrayDefinition()
        self.__worka5: GridArrayDefinition = GridArrayDefinition()
        self.__worka6: GridArrayDefinition = GridArrayDefinition()
        self.__worka7: GridArrayDefinition = GridArrayDefinition()
        self.__worka8: GridArrayDefinition = GridArrayDefinition()
        self.__worka9: GridArrayDefinition = GridArrayDefinition()
        self.__dx: GridArrayDefinition = GridArrayDefinition()
        self.__dy: GridArrayDefinition = GridArrayDefinition()
        self.__dz: GridArrayDefinition = GridArrayDefinition()
        self.__depth: GridArrayDefinition = GridArrayDefinition()
        self.__mdepth: GridArrayDefinition = GridArrayDefinition()
        self.__dznet: GridArrayDefinition = GridArrayDefinition()
        self.__compr: GridArrayDefinition = GridArrayDefinition()
        self.__icoars: GridArrayDefinition = GridArrayDefinition()
        self.__ialphaf: GridArrayDefinition = GridArrayDefinition()
        self.__ipolymer: GridArrayDefinition = GridArrayDefinition()
        self.__iadsorption: GridArrayDefinition = GridArrayDefinition()
        self.__itracer: GridArrayDefinition = GridArrayDefinition()
        self.__igrid: GridArrayDefinition = GridArrayDefinition()
        self.__isector: GridArrayDefinition = GridArrayDefinition()
        self.__swl: GridArrayDefinition = GridArrayDefinition()
        self.__swr: GridArrayDefinition = GridArrayDefinition()
        self.__swu: GridArrayDefinition = GridArrayDefinition()
        self.__sgl: GridArrayDefinition = GridArrayDefinition()
        self.__sgr: GridArrayDefinition = GridArrayDefinition()
        self.__sgu: GridArrayDefinition = GridArrayDefinition()
        self.__swro: GridArrayDefinition = GridArrayDefinition()
        self.__swro_ls: GridArrayDefinition = GridArrayDefinition()
        self.__sgro: GridArrayDefinition = GridArrayDefinition()
        self.__sgrw: GridArrayDefinition = GridArrayDefinition()
        self.__krw_swro: GridArrayDefinition = GridArrayDefinition()
        self.__krws_ls: GridArrayDefinition = GridArrayDefinition()
        self.__krw_swu: GridArrayDefinition = GridArrayDefinition()
        self.__krg_sgro: GridArrayDefinition = GridArrayDefinition()
        self.__krg_sgu: GridArrayDefinition = GridArrayDefinition()
        self.__krg_sgrw: GridArrayDefinition = GridArrayDefinition()
        self.__kro_swl: GridArrayDefinition = GridArrayDefinition()
        self.__kro_swr: GridArrayDefinition = GridArrayDefinition()
        self.__kro_sgl: GridArrayDefinition = GridArrayDefinition()
        self.__kro_sgr: GridArrayDefinition = GridArrayDefinition()
        self.__krw_sgl: GridArrayDefinition = GridArrayDefinition()
        self.__krw_sgr: GridArrayDefinition = GridArrayDefinition()
        self.__sgtr: GridArrayDefinition = GridArrayDefinition()
        self.__sotr: GridArrayDefinition = GridArrayDefinition()
        self.__swlpc: GridArrayDefinition = GridArrayDefinition()
        self.__sglpc: GridArrayDefinition = GridArrayDefinition()
        self.__pcw_swl: GridArrayDefinition = GridArrayDefinition()
        self.__pcg_sgu: GridArrayDefinition = GridArrayDefinition()
        self.__chloride: GridArrayDefinition = GridArrayDefinition()
        self.__calcium: GridArrayDefinition = GridArrayDefinition()
        self.__salinity: GridArrayDefinition = GridArrayDefinition()
        self.__api: GridArrayDefinition = GridArrayDefinition()
        self.__tmx: GridArrayDefinition = GridArrayDefinition()
        self.__tmy: GridArrayDefinition = GridArrayDefinition()
        self.__tmz: GridArrayDefinition = GridArrayDefinition()
        self.__multbv: GridArrayDefinition = GridArrayDefinition()
        self.__pv: GridArrayDefinition = GridArrayDefinition()

    def __wrap(self, value: Any) -> Any:
        if isinstance(value, tuple | list | set | frozenset):
            return type(value)([self.__wrap(v) for v in value])
        else:
            return value

    def update_properties_from_dict(self, data: dict[str, int | GridArrayDefinition]) -> None:
        """Allows you to update properties on the class using the provided dict of values.

        Args:
        ----
                data dict[str, int | GridArrayDefinition]: the dictionary of values to update on the class
        """
        # Use the dict provided to populate the properties in the class
        if data is not None:
            for name, value in data.items():
                private_name = '_' + name
                setattr(self, private_name, self.__wrap(value))

        # Prevent reload from disk
        self._grid_properties_loaded = True

    def to_dict(self) -> dict[str, Optional[int] | GridArrayDefinition]:
        self.load_grid_properties_if_not_loaded()

        original_dict = self.__dict__
        new_dict: dict[str, Optional[int] | GridArrayDefinition] = {}
        for key in original_dict.keys():
            new_key = key
            if new_key[0] == '_':
                new_key = new_key.replace('_', '', 1)
            if new_key[0] == '_':
                new_key = new_key.replace('_', '', 1)
            new_dict[new_key] = original_dict[key]

        return new_dict

    def load_grid_properties_if_not_loaded(self) -> None:
        def move_next_value(next_line: str) -> tuple[str, str]:
            """Finds the next value and then strips out the value from the line.

            Args:
            ----
                next_line (str): the line to search through for the value

            Raises:
            ------
                ValueError: if no value is found within the line provided

            Returns:
            -------
                tuple[str, str]: the next value found in the line, the line with the value stripped out.
            """
            value = nfo.get_next_value(0, [next_line], next_line)
            if value is None:
                raise ValueError(f"No value found within the provided line: {next_line}")
            next_line = next_line.replace(value, "", 1)
            return value, next_line

        # If we've already loaded the grid properties, don't do so again.
        if self._grid_properties_loaded:
            return

        if self.__grid_nexus_file is None or self.__grid_file_contents is None or self.__grid_file_nested is None:
            raise ValueError("Grid file not found, cannot load grid properties")

        # Strip file of comments
        file_as_list = nfo.strip_file_of_comments(self.__grid_file_contents, comment_characters=['!', 'C'])
        # Ignore blank lines
        file_as_list = [line for line in file_as_list if not line.strip() == '']

        properties_to_load = [
            PropertyToLoad('NETGRS', GRID_ARRAY_FORMAT_KEYWORDS, self._netgrs),
            PropertyToLoad('POROSITY', GRID_ARRAY_FORMAT_KEYWORDS, self._porosity),
            PropertyToLoad('POR', GRID_ARRAY_FORMAT_KEYWORDS, self._porosity),
            PropertyToLoad('SW', GRID_ARRAY_FORMAT_KEYWORDS, self._sw),
            PropertyToLoad('SG', GRID_ARRAY_FORMAT_KEYWORDS, self._sg),
            PropertyToLoad('PRESSURE', GRID_ARRAY_FORMAT_KEYWORDS, self._pressure),
            PropertyToLoad('P', GRID_ARRAY_FORMAT_KEYWORDS, self._pressure),
            PropertyToLoad('TEMPERATURE', GRID_ARRAY_FORMAT_KEYWORDS, self._temperature),
            PropertyToLoad('TEMP', GRID_ARRAY_FORMAT_KEYWORDS, self._temperature),
            PropertyToLoad('KX', GRID_ARRAY_FORMAT_KEYWORDS, self._kx),
            PropertyToLoad('KI', GRID_ARRAY_FORMAT_KEYWORDS, self._kx),
            PropertyToLoad('PERMX', GRID_ARRAY_FORMAT_KEYWORDS, self._kx),
            PropertyToLoad('PERMI', GRID_ARRAY_FORMAT_KEYWORDS, self._kx),
            PropertyToLoad('KY', GRID_ARRAY_FORMAT_KEYWORDS, self._ky),
            PropertyToLoad('KJ', GRID_ARRAY_FORMAT_KEYWORDS, self._ky),
            PropertyToLoad('PERMY', GRID_ARRAY_FORMAT_KEYWORDS, self._ky),
            PropertyToLoad('PERMJ', GRID_ARRAY_FORMAT_KEYWORDS, self._ky),
            PropertyToLoad('KZ', GRID_ARRAY_FORMAT_KEYWORDS, self._kz),
            PropertyToLoad('KK', GRID_ARRAY_FORMAT_KEYWORDS, self._kz),
            PropertyToLoad('PERMZ', GRID_ARRAY_FORMAT_KEYWORDS, self._kz),
            PropertyToLoad('PERMK', GRID_ARRAY_FORMAT_KEYWORDS, self._kz),
            PropertyToLoad('PVMULT', GRID_ARRAY_FORMAT_KEYWORDS, self.__pvmult),
            PropertyToLoad('CORP', ['VALUE'], self.__corp),
            PropertyToLoad('IEQUIL', GRID_ARRAY_FORMAT_KEYWORDS, self.__iequil),
            PropertyToLoad('IPVT', GRID_ARRAY_FORMAT_KEYWORDS, self.__ipvt),
            PropertyToLoad('IWATER', GRID_ARRAY_FORMAT_KEYWORDS, self.__iwater),
            PropertyToLoad('IRELPM', GRID_ARRAY_FORMAT_KEYWORDS, self.__irelpm),
            PropertyToLoad('IROCK', GRID_ARRAY_FORMAT_KEYWORDS, self.__irock),
            PropertyToLoad('ITRAN', GRID_ARRAY_FORMAT_KEYWORDS, self.__itran),
            PropertyToLoad('IREGION', GRID_ARRAY_FORMAT_KEYWORDS, self.__iregion),
            PropertyToLoad('LIVECELL', GRID_ARRAY_FORMAT_KEYWORDS, self.__livecell),
            PropertyToLoad('WORKA1', GRID_ARRAY_FORMAT_KEYWORDS, self.__worka1),
            PropertyToLoad('WORKA2', GRID_ARRAY_FORMAT_KEYWORDS, self.__worka2),
            PropertyToLoad('WORKA3', GRID_ARRAY_FORMAT_KEYWORDS, self.__worka3),
            PropertyToLoad('WORKA4', GRID_ARRAY_FORMAT_KEYWORDS, self.__worka4),
            PropertyToLoad('WORKA5', GRID_ARRAY_FORMAT_KEYWORDS, self.__worka5),
            PropertyToLoad('WORKA6', GRID_ARRAY_FORMAT_KEYWORDS, self.__worka6),
            PropertyToLoad('WORKA7', GRID_ARRAY_FORMAT_KEYWORDS, self.__worka7),
            PropertyToLoad('WORKA8', GRID_ARRAY_FORMAT_KEYWORDS, self.__worka8),
            PropertyToLoad('WORKA9', GRID_ARRAY_FORMAT_KEYWORDS, self.__worka9),
            PropertyToLoad('DX', GRID_ARRAY_FORMAT_KEYWORDS, self.__dx),
            PropertyToLoad('DY', GRID_ARRAY_FORMAT_KEYWORDS, self.__dy),
            PropertyToLoad('DZ', GRID_ARRAY_FORMAT_KEYWORDS, self.__dz),
            PropertyToLoad('DEPTH', GRID_ARRAY_FORMAT_KEYWORDS, self.__depth),
            PropertyToLoad('MDEPTH', GRID_ARRAY_FORMAT_KEYWORDS, self.__mdepth),
            PropertyToLoad('DZNET', GRID_ARRAY_FORMAT_KEYWORDS, self.__dznet),
            PropertyToLoad('COMPR', GRID_ARRAY_FORMAT_KEYWORDS, self.__compr),
            PropertyToLoad('CR', GRID_ARRAY_FORMAT_KEYWORDS, self.__compr),
            PropertyToLoad('ICOARS', GRID_ARRAY_FORMAT_KEYWORDS, self.__icoars),
            PropertyToLoad('IALPHAF', GRID_ARRAY_FORMAT_KEYWORDS, self.__ialphaf),
            PropertyToLoad('IPOLYMER', GRID_ARRAY_FORMAT_KEYWORDS, self.__ipolymer),
            PropertyToLoad('IADSORPTION', GRID_ARRAY_FORMAT_KEYWORDS, self.__iadsorption),
            PropertyToLoad('ITRACER', GRID_ARRAY_FORMAT_KEYWORDS, self.__itracer),
            PropertyToLoad('IGRID', GRID_ARRAY_FORMAT_KEYWORDS, self.__igrid),
            PropertyToLoad('ISECTOR', GRID_ARRAY_FORMAT_KEYWORDS, self.__isector),
            PropertyToLoad('SWL', GRID_ARRAY_FORMAT_KEYWORDS, self.__swl),
            PropertyToLoad('SWR', GRID_ARRAY_FORMAT_KEYWORDS, self.__swr),
            PropertyToLoad('SWU', GRID_ARRAY_FORMAT_KEYWORDS, self.__swu),
            PropertyToLoad('SGL', GRID_ARRAY_FORMAT_KEYWORDS, self.__sgl),
            PropertyToLoad('SGR', GRID_ARRAY_FORMAT_KEYWORDS, self.__sgr),
            PropertyToLoad('SGU', GRID_ARRAY_FORMAT_KEYWORDS, self.__sgu),
            PropertyToLoad('SWRO', GRID_ARRAY_FORMAT_KEYWORDS, self.__swro),
            PropertyToLoad('SWRO_LS', GRID_ARRAY_FORMAT_KEYWORDS, self.__swro_ls),
            PropertyToLoad('SGRO', GRID_ARRAY_FORMAT_KEYWORDS, self.__sgro),
            PropertyToLoad('SGRW', GRID_ARRAY_FORMAT_KEYWORDS, self.__sgrw),
            PropertyToLoad('KRW_SWRO', GRID_ARRAY_FORMAT_KEYWORDS, self.__krw_swro),
            PropertyToLoad('KRWRO', GRID_ARRAY_FORMAT_KEYWORDS, self.__krw_swro),
            PropertyToLoad('KRWS_LS', GRID_ARRAY_FORMAT_KEYWORDS, self.__krws_ls),
            PropertyToLoad('KRW_SWU', GRID_ARRAY_FORMAT_KEYWORDS, self.__krw_swu),
            PropertyToLoad('KRG_SGRO', GRID_ARRAY_FORMAT_KEYWORDS, self.__krg_sgro),
            PropertyToLoad('KRGRO', GRID_ARRAY_FORMAT_KEYWORDS, self.__krg_sgro),
            PropertyToLoad('KRG_SGU', GRID_ARRAY_FORMAT_KEYWORDS, self.__krg_sgu),
            PropertyToLoad('KRG_SGRW', GRID_ARRAY_FORMAT_KEYWORDS, self.__krg_sgrw),
            PropertyToLoad('KRGRW', GRID_ARRAY_FORMAT_KEYWORDS, self.__krg_sgrw),
            PropertyToLoad('KRO_SWL', GRID_ARRAY_FORMAT_KEYWORDS, self.__kro_swl),
            PropertyToLoad('KROLW', GRID_ARRAY_FORMAT_KEYWORDS, self.__kro_swl),
            PropertyToLoad('KRO_SWR', GRID_ARRAY_FORMAT_KEYWORDS, self.__kro_swr),
            PropertyToLoad('KRO_SGL', GRID_ARRAY_FORMAT_KEYWORDS, self.__kro_sgl),
            PropertyToLoad('KRO_SGR', GRID_ARRAY_FORMAT_KEYWORDS, self.__kro_sgr),
            PropertyToLoad('KRW_SGL', GRID_ARRAY_FORMAT_KEYWORDS, self.__krw_sgl),
            PropertyToLoad('KRW_SGR', GRID_ARRAY_FORMAT_KEYWORDS, self.__krw_sgr),
            PropertyToLoad('SGTR', GRID_ARRAY_FORMAT_KEYWORDS, self.__sgtr),
            PropertyToLoad('SOTR', GRID_ARRAY_FORMAT_KEYWORDS, self.__sotr),
            PropertyToLoad('SWLPC', GRID_ARRAY_FORMAT_KEYWORDS, self.__swlpc),
            PropertyToLoad('SGLPC', GRID_ARRAY_FORMAT_KEYWORDS, self.__sglpc),
            PropertyToLoad('PCW_SWL', GRID_ARRAY_FORMAT_KEYWORDS, self.__pcw_swl),
            PropertyToLoad('PCG_SGU', GRID_ARRAY_FORMAT_KEYWORDS, self.__pcg_sgu),
            PropertyToLoad('CHLORIDE', GRID_ARRAY_FORMAT_KEYWORDS, self.__chloride),
            PropertyToLoad('CALCIUM', GRID_ARRAY_FORMAT_KEYWORDS, self.__calcium),
            PropertyToLoad('SALINITY', GRID_ARRAY_FORMAT_KEYWORDS, self.__salinity),
            PropertyToLoad('SAL', GRID_ARRAY_FORMAT_KEYWORDS, self.__salinity),
            PropertyToLoad('API', GRID_ARRAY_FORMAT_KEYWORDS, self.__api),
            PropertyToLoad('TMX', GRID_ARRAY_FORMAT_KEYWORDS, self.__tmx),
            PropertyToLoad('TMY', GRID_ARRAY_FORMAT_KEYWORDS, self.__tmy),
            PropertyToLoad('TMZ', GRID_ARRAY_FORMAT_KEYWORDS, self.__tmz),
            PropertyToLoad('MULTBV', GRID_ARRAY_FORMAT_KEYWORDS, self.__multbv),
            PropertyToLoad('PV', GRID_ARRAY_FORMAT_KEYWORDS, self.__pv)
        ]

        possible_properties = [prop.token for prop in properties_to_load]

        for idx, line in enumerate(file_as_list):

            # Load in the basic properties
            line_start_token = nfo.get_next_value(0, [line])
            if line_start_token is not None and line_start_token.upper() in possible_properties:
                token_property = properties_to_load[possible_properties.index(line_start_token)]
                for modifier in token_property.modifiers:
                    # execute the load
                    StructuredGridOperations.load_token_value_if_present(
                        token_property.token, modifier, token_property.property, line, file_as_list, idx,
                        ['INCLUDE', 'NOLIST'])

            # Load in grid dimensions
            if nfo.check_token('NX', line):
                # Check that the format of the grid is NX followed by NY followed by NZ
                current_line = file_as_list[file_as_list.index(line)]
                remaining_line = current_line[current_line.index('NX') + 2:]
                if nfo.get_next_value(0, [remaining_line], remaining_line) != 'NY':
                    continue
                remaining_line = remaining_line[remaining_line.index('NY') + 2:]
                if nfo.get_next_value(0, [remaining_line], remaining_line) != 'NZ':
                    continue

                # Avoid loading in a comment
                if "!" in line and line.index("!") < line.index('NX'):
                    continue
                next_line = file_as_list[file_as_list.index(line) + 1]
                first_value, next_line = move_next_value(next_line)
                second_value, next_line = move_next_value(next_line)
                third_value, next_line = move_next_value(next_line)

                self._range_x = int(first_value)
                self._range_y = int(second_value)
                self._range_z = int(third_value)

        self._grid_properties_loaded = True

    @classmethod
    def load_structured_grid_file(cls: type[NexusGrid], structured_grid_file: File,
                                  lazy_loading: bool = True) -> NexusGrid:
        """Loads in a structured grid file with all grid properties, and the array functions defined with 'FUNCTION'.
        Other grid modifiers are currently not supported.

        Args:
        ----
            structured_grid_file (NexusFile): the NexusFile representation of a structured grid file for converting \
                into a structured grid file class
        Raises:
            AttributeError: if no value is found for the structured grid file path
            ValueError: if when loading the grid no values can be found for the NX NY NZ line.
        """
        if structured_grid_file.location is None:
            raise ValueError(f"No file path given or found for structured grid file path. \
                Instead got {structured_grid_file.location}")

        if not isinstance(structured_grid_file, NexusFile):
            raise ValueError(f"Cannot load file of type {type(structured_grid_file)}.")

        loaded_structured_grid_file = cls(grid_nexus_file=structured_grid_file)

        if not lazy_loading:
            loaded_structured_grid_file.load_grid_properties_if_not_loaded()
            loaded_structured_grid_file.load_faults()
            loaded_structured_grid_file.load_array_functions()

        return loaded_structured_grid_file

    @staticmethod
    def update_structured_grid_file(grid_dict: dict[str, GridArrayDefinition | int], model: NexusSimulator) -> None:
        """Save values passed from the front end to the structured grid file and update the class.

        Args:
        ----
            grid_dict (dict[str, Union[VariableEntry, int]]): dictionary containing grid properties to be replaced
            model (NexusSimulator): an instance of a NexusSimulator object
        Raises:
            ValueError: If no structured grid file is in the instance of the Simulator class
        """
        # Convert the dictionary back to a class, and update the properties on our class
        structured_grid = model.grid
        if structured_grid is None or model.model_files.structured_grid_file is None:
            raise ValueError("Model does not contain a structured grid")
        original_structured_grid_file = copy.deepcopy(structured_grid)

        # replace the structured grid with a new object with an updated dictionary
        structured_grid.update_properties_from_dict(grid_dict)

        # change it in the text file for nexus:
        grid_file_path = model.model_files.structured_grid_file.location
        if grid_file_path is None:
            raise ValueError("No path found for structured grid file path.")
        structured_grid_contents = nfo.load_file_as_list(grid_file_path)
        # Get the existing file as a list
        if structured_grid_contents is None:
            raise ValueError("No path found for structured grid file path. \
                Please provide a path to the structured grid")
        # Update each value in the file
        StructuredGridOperations.replace_value(structured_grid_contents, original_structured_grid_file.netgrs,
                                               structured_grid.netgrs, 'NETGRS')
        StructuredGridOperations.replace_value(structured_grid_contents, original_structured_grid_file.porosity,
                                               structured_grid.porosity, 'POROSITY')
        StructuredGridOperations.replace_value(structured_grid_contents, original_structured_grid_file.sw,
                                               structured_grid.sw, 'SW')
        StructuredGridOperations.replace_value(structured_grid_contents, original_structured_grid_file.kx,
                                               structured_grid.kx, 'KX')
        StructuredGridOperations.replace_value(structured_grid_contents, original_structured_grid_file.ky,
                                               structured_grid.ky, 'KY')
        StructuredGridOperations.replace_value(structured_grid_contents, original_structured_grid_file.kz,
                                               structured_grid.kz, 'KZ')

        # Save the new file contents
        new_file_str = "".join(structured_grid_contents)
        with open(grid_file_path, "w") as text_file:
            text_file.write(new_file_str)

    def load_array_functions(self) -> None:
        # for function arrays we need the expanded file contents without includes
        if self.__grid_nexus_file is None or self.__grid_nexus_file.get_flat_list_str_file is None:
            raise ValueError("Cannot load array functions as grid file cannot not found")
        file_contents = self.__grid_nexus_file.get_flat_list_str_file
        self.__array_functions_list = afo.collect_all_function_blocks(file_contents)
        self.__grid_array_functions = afo.create_grid_array_function_objects(self.__array_functions_list)
        self.__array_functions_df = afo.summarize_model_functions(self.__array_functions_list)
        self.__array_functions_loaded = True

    def get_array_functions_list(self) -> Optional[list[list[str]]]:
        """Returns the grid array functions as a list of function lines."""
        if not self.__array_functions_loaded:
            self.load_array_functions()
        return self.__array_functions_list

    def get_array_functions_df(self) -> Optional[pd.DataFrame]:
        """Returns the grid array functions as a dataframe."""
        if not self.__array_functions_loaded:
            self.load_array_functions()
        return self.__array_functions_df

    def load_faults(self) -> None:
        """Function to read faults in Nexus grid file defined using MULT and FNAME keywords."""
        file_content_as_list = self.__grid_file_contents
        if file_content_as_list is None:
            raise ValueError('Grid file contents have not been loaded')
        df = load_nexus_fault_mult_table_from_list(file_content_as_list)

        if not df.empty:
            # Ensure resulting dataframe has uppercase column names
            df.columns = [col.upper() for col in df.columns]

            # Check if any multfls have been used in grid file and update fault trans multipliers accordingly
            f_names = df['NAME'].unique()
            f_mults = [1.] * len(f_names)
            mult_dict = dict(zip(f_names, f_mults))
            for line in file_content_as_list:
                if nfo.check_token('MULTFL', line):
                    fname = str(nfo.get_expected_token_value(
                        'MULTFL', line, file_content_as_list,
                        custom_message=f'{line} does not have a fault name following MULTFL'))
                    if fname in df['NAME'].unique():
                        tmult = float(str(nfo.get_expected_token_value(
                            fname, line, file_content_as_list,
                            custom_message=f'MULTFL {fname} does not have a numerical tmult value')))
                        mult_dict[fname] *= tmult
            mult_df = pd.DataFrame.from_dict(mult_dict, orient='index').reset_index()
            mult_df.columns = ['NAME', 'TMULT']
            new_df = df.merge(mult_df, how='left', on='NAME', validate='many_to_one')
            new_df['MULT'] = new_df['MULT'] * new_df['TMULT']
            self.__faults_df = new_df.drop(['TMULT'], axis=1)
        self.__grid_faults_loaded = True

    def get_faults_df(self) -> Optional[pd.DataFrame]:
        """Returns the fault definition and transmissility multiplier information as a dataframe."""
        self.load_grid_properties_if_not_loaded()
        if not self.__grid_faults_loaded:
            self.load_faults()
        return self.__faults_df

    @staticmethod
    def __keyword_in_include_file_warning(var_entry_obj: GridArrayDefinition) -> None:

        if var_entry_obj.keyword_in_include_file is True:
            warnings.warn('Grid array keyword in include file. This is not recommended simulation practice.')
        else:
            return None

    @property
    def array_functions(self) -> Optional[list[NexusGridArrayFunction]]:
        """Returns a list of the array functions defined in the structured grid file."""
        self.load_grid_properties_if_not_loaded()
        if self.__grid_array_functions is None:
            self.load_array_functions()
        return self.__grid_array_functions

    @property
    def corp(self) -> GridArrayDefinition:
        self.load_grid_properties_if_not_loaded()
        NexusGrid.__keyword_in_include_file_warning(self.__corp)

        return self.__corp

    @property
    def iequil(self) -> GridArrayDefinition:
        self.load_grid_properties_if_not_loaded()
        return self.__iequil

    @property
    def ipvt(self) -> GridArrayDefinition:
        self.load_grid_properties_if_not_loaded()
        return self.__ipvt

    @property
    def iwater(self) -> GridArrayDefinition:
        self.load_grid_properties_if_not_loaded()
        return self.__iwater

    @property
    def irelpm(self) -> GridArrayDefinition:
        self.load_grid_properties_if_not_loaded()
        return self.__irelpm

    @property
    def irock(self) -> GridArrayDefinition:
        self.load_grid_properties_if_not_loaded()
        return self.__irock

    @property
    def itran(self) -> GridArrayDefinition:
        self.load_grid_properties_if_not_loaded()
        return self.__itran

    @property
    def iregion(self) -> dict[str, GridArrayDefinition]:
        self.load_grid_properties_if_not_loaded()
        return self.__iregion

    @property
    def livecell(self) -> GridArrayDefinition:
        self.load_grid_properties_if_not_loaded()
        return self.__livecell

    @property
    def pvmult(self) -> GridArrayDefinition:
        self.load_grid_properties_if_not_loaded()
        return self.__pvmult

    @property
    def worka1(self) -> GridArrayDefinition:
        self.load_grid_properties_if_not_loaded()
        return self.__worka1

    @property
    def worka2(self) -> GridArrayDefinition:
        self.load_grid_properties_if_not_loaded()
        return self.__worka2

    @property
    def worka3(self) -> GridArrayDefinition:
        self.load_grid_properties_if_not_loaded()
        return self.__worka3

    @property
    def worka4(self) -> GridArrayDefinition:
        self.load_grid_properties_if_not_loaded()
        return self.__worka4

    @property
    def worka5(self) -> GridArrayDefinition:
        self.load_grid_properties_if_not_loaded()
        return self.__worka5

    @property
    def worka6(self) -> GridArrayDefinition:
        self.load_grid_properties_if_not_loaded()
        return self.__worka6

    @property
    def worka7(self) -> GridArrayDefinition:
        self.load_grid_properties_if_not_loaded()
        return self.__worka7

    @property
    def worka8(self) -> GridArrayDefinition:
        self.load_grid_properties_if_not_loaded()
        return self.__worka8

    @property
    def worka9(self) -> GridArrayDefinition:
        self.load_grid_properties_if_not_loaded()
        return self.__worka9

    @property
    def dx(self) -> GridArrayDefinition:
        self.load_grid_properties_if_not_loaded()
        return self.__dx

    @property
    def dy(self) -> GridArrayDefinition:
        self.load_grid_properties_if_not_loaded()
        return self.__dy

    @property
    def dz(self) -> GridArrayDefinition:
        self.load_grid_properties_if_not_loaded()
        return self.__dz

    @property
    def depth(self) -> GridArrayDefinition:
        self.load_grid_properties_if_not_loaded()
        return self.__depth

    @property
    def mdepth(self) -> GridArrayDefinition:
        self.load_grid_properties_if_not_loaded()
        return self.__mdepth

    @property
    def dznet(self) -> GridArrayDefinition:
        self.load_grid_properties_if_not_loaded()
        return self.__dznet

    @property
    def compr(self) -> GridArrayDefinition:
        self.load_grid_properties_if_not_loaded()
        return self.__compr

    @property
    def icoars(self) -> GridArrayDefinition:
        self.load_grid_properties_if_not_loaded()
        return self.__icoars

    @property
    def ialphaf(self) -> GridArrayDefinition:
        self.load_grid_properties_if_not_loaded()
        return self.__ialphaf

    @property
    def ipolymer(self) -> GridArrayDefinition:
        self.load_grid_properties_if_not_loaded()
        return self.__ipolymer

    @property
    def iadsorption(self) -> GridArrayDefinition:
        self.load_grid_properties_if_not_loaded()
        return self.__iadsorption

    @property
    def itracer(self) -> GridArrayDefinition:
        self.load_grid_properties_if_not_loaded()
        return self.__itracer

    @property
    def igrid(self) -> GridArrayDefinition:
        self.load_grid_properties_if_not_loaded()
        return self.__igrid

    @property
    def isector(self) -> GridArrayDefinition:
        self.load_grid_properties_if_not_loaded()
        return self.__isector

    @property
    def swl(self) -> GridArrayDefinition:
        self.load_grid_properties_if_not_loaded()
        return self.__swl

    @property
    def swr(self) -> GridArrayDefinition:
        self.load_grid_properties_if_not_loaded()
        return self.__swr

    @property
    def swu(self) -> GridArrayDefinition:
        self.load_grid_properties_if_not_loaded()
        return self.__swu

    @property
    def sgl(self) -> GridArrayDefinition:
        self.load_grid_properties_if_not_loaded()
        return self.__sgl

    @property
    def sgr(self) -> GridArrayDefinition:
        self.load_grid_properties_if_not_loaded()
        return self.__sgr

    @property
    def sgu(self) -> GridArrayDefinition:
        self.load_grid_properties_if_not_loaded()
        return self.__sgu

    @property
    def swro(self) -> GridArrayDefinition:
        self.load_grid_properties_if_not_loaded()
        return self.__swro

    @property
    def swro_ls(self) -> GridArrayDefinition:
        self.load_grid_properties_if_not_loaded()
        return self.__swro_ls

    @property
    def sgro(self) -> GridArrayDefinition:
        self.load_grid_properties_if_not_loaded()
        return self.__sgro

    @property
    def sgrw(self) -> GridArrayDefinition:
        self.load_grid_properties_if_not_loaded()
        return self.__sgrw

    @property
    def krw_swro(self) -> GridArrayDefinition:
        self.load_grid_properties_if_not_loaded()
        return self.__krw_swro

    @property
    def krws_ls(self) -> GridArrayDefinition:
        self.load_grid_properties_if_not_loaded()
        return self.__krws_ls

    @property
    def krw_swu(self) -> GridArrayDefinition:
        self.load_grid_properties_if_not_loaded()
        return self.__krw_swu

    @property
    def krg_sgro(self) -> GridArrayDefinition:
        self.load_grid_properties_if_not_loaded()
        return self.__krg_sgro

    @property
    def krg_sgu(self) -> GridArrayDefinition:
        self.load_grid_properties_if_not_loaded()
        return self.__krg_sgu

    @property
    def krg_sgrw(self) -> GridArrayDefinition:
        self.load_grid_properties_if_not_loaded()
        return self.__krg_sgrw

    @property
    def kro_swl(self) -> GridArrayDefinition:
        self.load_grid_properties_if_not_loaded()
        return self.__kro_swl

    @property
    def kro_swr(self) -> GridArrayDefinition:
        self.load_grid_properties_if_not_loaded()
        return self.__kro_swr

    @property
    def kro_sgl(self) -> GridArrayDefinition:
        self.load_grid_properties_if_not_loaded()
        return self.__kro_sgl

    @property
    def kro_sgr(self) -> GridArrayDefinition:
        self.load_grid_properties_if_not_loaded()
        return self.__kro_sgr

    @property
    def krw_sgl(self) -> GridArrayDefinition:
        self.load_grid_properties_if_not_loaded()
        return self.__krw_sgl

    @property
    def krw_sgr(self) -> GridArrayDefinition:
        self.load_grid_properties_if_not_loaded()
        return self.__krw_sgr

    @property
    def sgtr(self) -> GridArrayDefinition:
        self.load_grid_properties_if_not_loaded()
        return self.__sgtr

    @property
    def sotr(self) -> GridArrayDefinition:
        self.load_grid_properties_if_not_loaded()
        return self.__sotr

    @property
    def swlpc(self) -> GridArrayDefinition:
        self.load_grid_properties_if_not_loaded()
        return self.__swlpc

    @property
    def sglpc(self) -> GridArrayDefinition:
        self.load_grid_properties_if_not_loaded()
        return self.__sglpc

    @property
    def pcw_swl(self) -> GridArrayDefinition:
        self.load_grid_properties_if_not_loaded()
        return self.__pcw_swl

    @property
    def pcg_sgu(self) -> GridArrayDefinition:
        self.load_grid_properties_if_not_loaded()
        return self.__pcg_sgu

    @property
    def chloride(self) -> GridArrayDefinition:
        self.load_grid_properties_if_not_loaded()
        return self.__chloride

    @property
    def calcium(self) -> GridArrayDefinition:
        self.load_grid_properties_if_not_loaded()
        return self.__calcium

    @property
    def salinity(self) -> GridArrayDefinition:
        self.load_grid_properties_if_not_loaded()
        return self.__salinity

    @property
    def api(self) -> GridArrayDefinition:
        self.load_grid_properties_if_not_loaded()
        return self.__api

    @property
    def tmx(self) -> GridArrayDefinition:
        self.load_grid_properties_if_not_loaded()
        return self.__tmx

    @property
    def tmy(self) -> GridArrayDefinition:
        self.load_grid_properties_if_not_loaded()
        return self.__tmy

    @property
    def tmz(self) -> GridArrayDefinition:
        self.load_grid_properties_if_not_loaded()
        return self.__tmz

    @property
    def multbv(self) -> GridArrayDefinition:
        self.load_grid_properties_if_not_loaded()
        return self.__multbv

    @property
    def pv(self) -> GridArrayDefinition:
        self.load_grid_properties_if_not_loaded()
        return self.__pv



from dataclasses import dataclass
from typing import Optional

from ResSimpy.Completion import Completion
from ResSimpy.Enums.PenetrationDirectionEnum import PenetrationDirectionEnum
from ResSimpy.Nexus.NexusEnums.DateFormatEnum import DateFormat
from ResSimpy.Units.AttributeMappings.CompletionUnitMapping import CompletionUnits


@dataclass(kw_only=True)
class OpenGoSimCompletion(Completion):

    __penetration_direction: Optional[PenetrationDirectionEnum]
    __is_open: bool
    __refinement_name: Optional[str]

    def __init__(self, date: str, i: Optional[int] = None, j: Optional[int] = None, k: Optional[int] = None,
                 penetration_direction: Optional[PenetrationDirectionEnum] = None, is_open: Optional[bool] = None,
                 refinement_name: Optional[str] = None) -> None:
        """Initialises the OpenGoSimCompletion class.

        Args:
            date (str): The date of the completion.
            i (Optional[int]): The i index of the completion.
            j (Optional[int]): The j index of the completion.
            k (Optional[int]): The k index of the completion.
            penetration_direction (Optional[PenetrationDirectionEnum]): The direction of the penetration for the
            completion.
            is_open (Optional[bool]): Whether the completion is open or closed. If True the completion is open.
            refinement_name (Optional[str]): The grid refinement name that the completion is applied to.
        """
        self.__penetration_direction = penetration_direction
        self.__is_open = is_open if is_open is not None else True
        self.__refinement_name = refinement_name
        super().__init__(i=i, j=j, k=k, date=date, date_format=DateFormat.DD_MMM_YYYY)

    def __repr__(self) -> str:
        return f"i: {self.i} j: {self.j} k: {self.k}"

    def is_open_set(self, value: bool) -> None:
        """Set the open state of a completion."""
        self.__is_open = value

    @property
    def refinement_name(self) -> Optional[str]:
        """The refinement name that the completion is applied to."""
        return self.__refinement_name

    @property
    def penetration_direction(self) -> Optional[PenetrationDirectionEnum]:
        """The direction of the penetration for the completion."""
        return self.__penetration_direction

    @property
    def is_open(self) -> bool:
        """Whether the completion is perforated or not."""
        return self.__is_open

    @staticmethod
    def get_keyword_mapping() -> dict[str, tuple[str, type]]:
        """Gets the mapping of keywords to attribute definitions."""
        raise NotImplementedError("Not implemented for OGS yet")

    @property
    def units(self) -> CompletionUnits:
        """Returns the attribute to unit map for the data object."""
        raise NotImplementedError("Not implemented for OGS yet")

    @property
    def completion_is_perforation(self) -> bool:
        """Determines if the supplied completion is a perforation or not."""
        return self.__is_open



"""Functions for loading a well list from a surface network file."""

from ResSimpy.Nexus.DataModels.NexusWellList import NexusWellList
import ResSimpy.Nexus.nexus_file_operations as nfo
from ResSimpy.Enums.HowEnum import OperationEnum


def load_well_lists(file_as_list: list[str], current_date: None | str,
                    previous_well_lists: None | list[NexusWellList] = None,
                    table_start_index: int = 0) -> list[tuple[NexusWellList, int]]:
    """Loads a well list from a surface network file.

    Args:
        file_as_list (list[str]): The surface network file as a list of strings.
        current_date (None | str): The current date of the simulation.
        previous_well_lists (None | list[tuple[NexusWellList, int]]):
        The list of tuples of already loaded in NexusWellLists with their starting index. Defaults to None.
        table_start_index (int): The starting index of the table. Defaults to 0.

    Returns:
        NexusWellList: A NexusWellList object containing the well list data.
    """
    if previous_well_lists is None:
        previous_well_lists = []
    if current_date is None:
        current_date = ''

    well_list_name = nfo.get_expected_token_value('WELLLIST', file_as_list[0], file_as_list)

    previous_well_list = get_previous_well_list(well_list_name=well_list_name, well_lists=previous_well_lists)

    new_well_list = load_well_list_from_table(well_list_as_list_str=file_as_list, current_date=current_date,
                                              previous_well_list=previous_well_list, well_list_name=well_list_name)

    return [(new_well_list, table_start_index)]


def get_previous_well_list(well_list_name: str, well_lists: list[NexusWellList]) -> NexusWellList | None:
    """Gets the previous well list from a list of well lists.

    Args:
        well_list_name (str): The name of the well list to get.
        well_lists (list[NexusWellList]): The list of well lists to search.

    Returns:
        NexusWellList | None: The previous well list if it exists, None otherwise.
    """
    for well_list in well_lists:
        if well_list.name == well_list_name:
            return well_list
    return None


def load_well_list_from_table(well_list_as_list_str: list[str], current_date: str, well_list_name: str,
                              previous_well_list: NexusWellList | None = None) -> NexusWellList:
    """Loads a well list from a table taken from a surface network file.
    Currently does not support Wildcards in welllists.

    Args:
        well_list_as_list_str (list[str]): The subsection of surface network file as a list of strings.
        current_date (str): The current date of the simulation.
        previous_well_list (NexusWellList | None): The previous well list. Defaults to None.

    Returns:
        NexusWellList: A NexusWellList object containing the well list data.
    """
    well_list = previous_well_list.wells.copy() if previous_well_list is not None else []
    operation: OperationEnum | None = None
    for line in well_list_as_list_str:
        if nfo.check_token('CLEAR', line):
            # clear the existing welllist
            well_list = []
            continue

        if nfo.check_token('ADD', line):
            operation = OperationEnum.ADD
            continue
        elif nfo.check_token('REMOVE', line):
            operation = OperationEnum.REMOVE
            continue

        if operation is None:
            continue
        if nfo.check_token('ENDWELLLIST', line):
            break

        # otherwise the well should be removed/added from/to the welllist.
        well_name = nfo.get_next_value(0, [line])
        if well_name is None:
            # handle the empty line case
            continue

        if operation == OperationEnum.ADD:
            well_list.append(well_name)
        elif operation == OperationEnum.REMOVE:
            if well_name in well_list:
                well_list.remove(well_name)

    # TODO may also need to store the line number for the loaded well list
    return NexusWellList(name=well_list_name, wells=well_list, date=current_date)



from dataclasses import dataclass

from ResSimpy.Network import Network


@dataclass(kw_only=True, init=False)
class OpenGoSimNetwork(Network):
    # Class to be implemented later

    def __init__(self) -> None:
        """Initialises the OpenGoSimNetwork class.

        Args:
            To be implemented later.
        """
        self.nodes = None
        self.connections = None
        self.constraints = None
        self.targets = None

    def __repr__(self) -> str:
        # TODO: implement this
        return "Not implemented yet"

    def load(self) -> None:
        raise NotImplementedError("Not implemented for OGS yet")



# TODO: complete these lists

GENERAL_KEYWORDS = ['DATE', ' TIME', 'END']

# TODO: GRDECL KEYWORDS E.G DX, PERMX, ZCORN ETC

GRID_KEYWORDS = ['TYPE']

SIMULATION_KEYWORDS = ['MODE', 'ISOTHERMAL', 'NOSLVSOL', 'NOGASSOL', 'RESERVOIR_DEFAULTS', 'ANALYTICAL_JACOBIAN',
                       'HYSTERESIS', 'HYSTERESIS_PC', 'EHYST', 'RESTART', 'ERESTART', 'CHECKPOINT', 'TL_OMEGA',
                       'COMP_DEP_SFNS', 'STRAND', 'ABOUT', 'PREAMBLE', 'FLOWS', 'SATDEN', 'OPTIONS', 'MISCIBLE',
                       'LOCAL_DIAGONAL_TENSOR']

OUTPUT_KEYWORDS = ['MASS_BALANCE_FILE', 'PERIODIC TIMESTEP', 'ECLIPSE_FILE', 'PERIOD_SUM TIMESTEP',
                   'PERIOD_RST TIMESTEP', 'WRITE_DENSITY', 'LINEREPT', 'SOLUTION_MONITOR', 'TIME_UNITS', 'OUTFILE']

MATERIAL_PROP_KEYWORDS = ['ID', 'CHARACTERISTIC_CURVES', 'ROCK_COMPRESSIBILITY', 'ROCK_REFERENCE_PRESSURE',
                          'ROCK_COMPRESSIBILITY_FUNCTION', 'ROCK_DENSITY', 'SPECIFIC_HEAT', 'THERMAL_CONDUCTIVITY_DRY',
                          'THERMAL_CONDUCTIVITY_WET', 'SOIL_COMPRESSIBIITY', 'SOIL_REFERENCE_PRESSURE', 'TORTUOSITY']

# TODO: ANALYTIC APPROACH KEYWORDS ALTHOUGH TABLES IS PROBABLY MOST COMMON COVERED BY BELOW

SAT_FUNCTIONS_KEYWORDS = ['TABLE', 'PRESSURE_UNITS', 'SWFN', 'SGFN', 'SGFN', 'SGSFN', 'SOF2', 'SOF3']

EOS_KEYWORDS = ['WATER', 'OIL', 'GAS', 'SOLVENT', 'COMP', 'SURFACE_DENSITY', 'FORMULA_WEIGHT', 'DATABASE',
                'PREOS', 'SRKEOS', 'CNAME', 'TCRIT', 'PCRIT', 'ZCRIT', 'ZMF', 'ACF', 'PRCORR', 'MW', 'SHIFT',
                'OMEGAA', 'OMEGAB', 'BIC', 'GCOND', 'PARACHOR']

WELLS_KEYWORDS = ['WELL_TYPE', 'OPEN', 'SHUT', 'RADIUS', 'DIAMETER', 'USE_HIST_TARG', 'CIJK_D', 'CIJKL_D',
                  'SKIN_FACTOR', 'EFFECTIVE_KH', 'PRESSURE_RADIUS_R0', 'CONST_DRILL_DIR', 'BHPL', 'D_REF',
                  'Z_REF', 'INJECTION_ENTHALPY_P', 'INJECTION_ENTHALPY_T', 'ZINJ', 'MULTI_PI', 'GC_WEIGHT']

AQUIFER_KEYWORDS = ['BACKFLOW', 'DEPTH', 'ZAQUIFER', 'THETA_FRACTION', 'THETA', 'THICKNESS', 'WIDTH', 'RADIUS',
                    'PERM', 'PERMEABILITY', 'CMPR', 'COMPRESSIBILITY', 'PORO', 'POROSITY', 'VISC', 'VISCOSITY',
                    'PINIT', 'PRESSURE', 'TINIT', 'TEMPERATURE', 'GAS_IN_LIQUID_MOLE_FRACTION', 'IFT',
                    'CONN_D', 'CONN_Z', 'FET', 'PV', 'PI']

EQUILIBRATION_KEYWORDS = ['DATUM_D', 'PRESSURE', 'WGC_D', 'PCWG_WGC', 'OWC_D', 'PCOW_OWC', 'OGC_D', 'PCOG_OGC',
                          'BUBBLE_POINT_TABLE', 'RTEMP', 'TEMPERATURE_TABLE', 'SALT_TABLE',
                          'GAS_IN_LIQUID_MOLE_FRACTION', 'SOLVENT_IN_LIQUID_MOLE_FRACTION', 'ZMFVD',
                          'NUM_EQUIL_NODES']

OPENGOSIM_KEYWORDS = GENERAL_KEYWORDS + SIMULATION_KEYWORDS + OUTPUT_KEYWORDS + MATERIAL_PROP_KEYWORDS + \
                     SAT_FUNCTIONS_KEYWORDS + EOS_KEYWORDS + WELLS_KEYWORDS + AQUIFER_KEYWORDS + \
                     EQUILIBRATION_KEYWORDS



from dataclasses import dataclass

from ResSimpy.OpenGoSim.DataModels.OpenGoSimWell import OpenGoSimWell
from ResSimpy.Wells import Wells


@dataclass(kw_only=True, init=False)
class OpenGoSimWells(Wells):
    """A class for collecting all OpenGoSimWell objects in an OpenGoSimSimulator. Handles adding and removing
    completions as well as rescheduling wells. This is usually accessed through the model.wells property.
    """
    _wells: list[OpenGoSimWell]



from dataclasses import dataclass
from typing import Optional

import ResSimpy.FileOperations.file_operations as fo
from ResSimpy.Aquifer import Aquifer
from ResSimpy.Enums.PenetrationDirectionEnum import PenetrationDirectionEnum
from ResSimpy.Enums.UnitsEnum import UnitSystem
from ResSimpy.Enums.WellTypeEnum import WellType
from ResSimpy.Equilibration import Equilibration
from ResSimpy.File import File
from ResSimpy.Gaslift import Gaslift
from ResSimpy.Hydraulics import Hydraulics
from ResSimpy.OpenGoSim.DataModels.OpenGoSimCompletion import OpenGoSimCompletion
from ResSimpy.OpenGoSim.DataModels.OpenGoSimWell import OpenGoSimWell
from ResSimpy.OpenGoSim.Enums.SimulationTypeEnum import SimulationType
from ResSimpy.OpenGoSim.Model_Parts.OpenGoSimNetwork import OpenGoSimNetwork
from ResSimpy.OpenGoSim.OpenGoSimKeywords.OpenGoSimKeywords import OPENGOSIM_KEYWORDS
from ResSimpy.OpenGoSim.OpenGoSimWells import OpenGoSimWells
from ResSimpy.PVT import PVT
from ResSimpy.RelPerm import RelPerm
from ResSimpy.Reporting import Reporting
from ResSimpy.Rock import Rock
from ResSimpy.Separator import Separator
from ResSimpy.Simulator import Simulator
from ResSimpy.Valve import Valve
from ResSimpy.Water import Water


def line_contains_block_ending(line: str) -> bool:
    """Checks if a line contains a block ending or not."""
    if '/' in line or fo.check_token('END', line):
        return True
    return False


@dataclass(kw_only=True)
class OpenGoSimSimulator(Simulator):
    __simulation_type: SimulationType
    __final_date: Optional[str]

    def __init__(self, origin: str) -> None:
        """Initialises the OpenGoSimSimulator class.

        Args:
            origin (str): The path to the model file.
        """
        super().__init__()
        self._origin = origin
        self._wells: OpenGoSimWells = OpenGoSimWells()

        self.__final_date = None

        # Model parts not implemented yet. Will be replaced by OGS specific classes
        self._pvt = PVT()
        self._separator = Separator()
        self._water = Water()
        self._equil = Equilibration()
        self._rock = Rock()
        self._relperm = RelPerm()
        self._valve = Valve()
        self._aquifer = Aquifer()
        self._hydraulics = Hydraulics()
        self._gaslift = Gaslift()
        self._network = OpenGoSimNetwork()
        self._grid = None
        self._model_files = File(location=origin)
        self._default_units = UnitSystem.UNDEFINED
        self._reporting = Reporting(self)
        self.__load_model()

    @property
    def simulation_type(self) -> SimulationType:
        return self.__simulation_type

    @property
    def final_date(self) -> Optional[str]:
        return self.__final_date

    def __repr__(self) -> str:
        full_string = f"""Simulation Type {self.simulation_type.value}
Start Date: {self.start_date}
End Date: {self.final_date}

WELLS
-----
{self.wells.get_wells_overview()}
"""
        return full_string

    def __load_model(self) -> None:
        """Loads the full OGS model."""
        block_tokens = ['SIMULATION', 'WELL_DATA', 'TIME']
        model_as_list = fo.load_file_as_list(self._origin)

        for index, line in enumerate(model_as_list):
            for token in block_tokens:
                if fo.check_token(token=token, line=line):
                    # Begin loading in the 'block' of information
                    if token.upper() == 'SIMULATION':
                        self.__load_in_simulation_block(model_as_list[index:])

                    if token.upper() == 'WELL_DATA':
                        self.__load_in_well_data_block(model_as_list[index:])

                    if token.upper() == 'TIME':
                        self.__load_in_time_block(model_as_list[index:])

    def __load_in_time_block(self, remaining_text: list[str]) -> None:
        # Load in the TIME block of the model.
        for index, line in enumerate(remaining_text):
            if line_contains_block_ending(line):
                break
            if fo.check_token('START_DATE', line):
                self._start_date = fo.load_in_three_part_date(initial_token='START_DATE', token_line=line,
                                                              file_as_list=remaining_text, start_index=index)

            if fo.check_token('FINAL_DATE', line):
                self.__final_date = fo.load_in_three_part_date(initial_token='FINAL_DATE', token_line=line,
                                                               file_as_list=remaining_text, start_index=index)

    def __load_in_simulation_block(self, remaining_text: list[str]) -> None:
        # Load in the SIMULATION block of the model.
        for line in remaining_text:
            if line_contains_block_ending(line):
                break
            if fo.check_token('SIMULATION_TYPE', line):
                value = fo.get_expected_token_value(token='SIMULATION_TYPE', token_line=line, file_list=remaining_text)
                self.__simulation_type = SimulationType.SUBSURFACE if value.upper() == 'SUBSURFACE' \
                    else SimulationType.GEOMECHANICS_SUBSURFACE

    def __load_in_well_data_block(self, remaining_text: list[str]) -> None:
        # Load in the WELL_DATA block of the model.
        completions_to_add: list[OpenGoSimCompletion] = []
        unique_completions: list[OpenGoSimCompletion] = []
        well_name: str = fo.get_expected_token_value(token='WELL_DATA',
                                                     token_line=remaining_text[0],
                                                     file_list=remaining_text)
        well_type: WellType | None = None
        relevant_date = self.start_date
        open_status = True
        penetration_direction = PenetrationDirectionEnum.Z
        refinement_name = None
        for index, line in enumerate(remaining_text):
            if line_contains_block_ending(line):
                break
            if fo.check_token('WELL_TYPE', line):
                value = fo.get_expected_token_value(token='WELL_TYPE', token_line=line, file_list=remaining_text)
                well_type = WellType[value.upper()]
            if fo.check_token('DATE', line):
                relevant_date = fo.load_in_three_part_date(initial_token='DATE', token_line=line,
                                                           file_as_list=remaining_text, start_index=index)

                new_completions_to_add: list[OpenGoSimCompletion] = []

                # Make a copy of each existing completion for the new date
                for unique_completion in unique_completions:
                    new_completion = OpenGoSimCompletion(i=unique_completion.i, j=unique_completion.j,
                                                         k=unique_completion.k, date=relevant_date,
                                                         penetration_direction=unique_completion.penetration_direction,
                                                         is_open=open_status, refinement_name=refinement_name)
                    new_completions_to_add.append(new_completion)
                completions_to_add.extend(new_completions_to_add)
            if fo.check_token('OPEN', line):
                # Apply 'open' state to all previous completions on this date
                for completion in completions_to_add:
                    if completion.date == relevant_date:
                        completion.is_open_set(True)
            if fo.check_token('SHUT', line):
                # Apply 'shut' state to all previous completions on this date
                for completion in completions_to_add:
                    if completion.date == relevant_date:
                        completion.is_open_set(False)
            if fo.check_token('CIJK_D', line) or fo.check_token('CIJKL_D', line):
                # Load in the completions
                remaining_text_from_here = remaining_text[index:]

                values_in_order = fo.get_multiple_expected_sequential_values(list_of_strings=remaining_text_from_here,
                                                                             number_tokens=4,
                                                                             ignore_values=['CIJK_D', 'CIJKL_D'])

                i_value = values_in_order[0]
                j_value = values_in_order[1]
                k_1 = values_in_order[2]
                k_2 = values_in_order[3]

                # Get the refinement name and / or penetration direction
                next_value = fo.get_nth_value(list_of_strings=remaining_text_from_here, value_number=5,
                                              ignore_values=['CIJK_D', 'CIJKL_D'])

                if fo.check_token('CIJKL_D', line):
                    # Completion is in grid refinement, get the name of that refinement first
                    refinement_name = next_value
                    next_value = fo.get_nth_value(list_of_strings=remaining_text_from_here, value_number=6,
                                                  ignore_values=['CIJKL_D'])

                # The next value, if present, will be the penetration direction.
                if next_value is not None and next_value.upper() not in OPENGOSIM_KEYWORDS:
                    penetration_direction = PenetrationDirectionEnum[next_value]

                # Create a new completion for every layer found, and add it to the well
                for c in range(int(k_1), int(k_2) + 1):
                    new_completion_to_add = OpenGoSimCompletion(i=int(i_value), j=int(j_value), k=c, date=relevant_date,
                                                                penetration_direction=penetration_direction,
                                                                refinement_name=refinement_name, is_open=open_status)
                    completions_to_add.append(new_completion_to_add)

                unique_completions = completions_to_add.copy()

        if well_type is None:
            raise ValueError(f"Cannot determine well type for well {well_name}")

        loaded_well = OpenGoSimWell(well_type=well_type, well_name=well_name, completions=completions_to_add)

        self._wells._wells.append(loaded_well)
        self._wells._wells_loaded = True

    @staticmethod
    def get_fluid_type(surface_file_content: list[str]) -> str:
        raise NotImplementedError("Not implemented for OGS yet")

    def set_output_path(self, path: str) -> None:
        raise NotImplementedError("Not implemented for OGS yet")

    def get_date_format(self) -> str:
        """Returns date format as a string."""
        raise NotImplementedError("Not implemented for OGS yet")

    def write_out_new_model(self, new_location: str, new_model_name: str) -> None:
        """Writes out a new version of the model to the location supplied.

        Args:
        new_location (str): Path to write the contents of the model to.
        """
        raise NotImplementedError("Not implemented for OGS yet")



"""Collection of Enumerations for well type used in ResSimpy."""
from enum import Enum


class WellType(str, Enum):
    """An enum representing the type of the well."""
    PRODUCER = 'PRODUCER'
    OIL_INJECTOR = 'OIL_INJECTOR'
    GAS_INJECTOR = 'GAS_INJECTOR'
    WATER_INJECTOR = 'WATER_INJECTOR'
    SOLVENT_INJECTOR = 'SOLVENT_INJECTOR'
    WAG_INJECTOR = 'WAG_INJECTOR'



"""Class to represent Nexus grid array functions."""
from dataclasses import dataclass
from typing import Optional
import pandas as pd

from ResSimpy.GridArrayFunction import GridArrayFunction


@dataclass
class NexusGridArrayFunction(GridArrayFunction):
    """Class to hold Nexus grid array function attributes.

    Attributes:
        function_table (pd.DataFrame): Dependency between input and output grid arrays, represented in tabular form
        function_table_m (int): Maximum number of function table entries which can be used in tabular interpolation
        function_table_p_list (list[float]): Shifts of the first, second, ..., and N-th input arrays
    """

    function_table: Optional[pd.DataFrame] = None
    function_table_m: Optional[int] = None
    function_table_p_list: Optional[list[float]] = None



"""Holds a class for representing a single WellList or group for the Nexus model."""
from dataclasses import dataclass

from ResSimpy.Units.AttributeMappings.BaseUnitMapping import BaseUnitMapping
from ResSimpy.WellLists import WellList


@dataclass(kw_only=True, init=False)
class NexusWellList(WellList):
    """Class for representing a single WellList or group for the Nexus model."""

    def __init__(self, name: str, wells: list[str], date: str) -> None:
        """Initialises the NexusWellList class.

        Args:
            name (str): Name of the welllist.
            wells (list[str]): List of well names in the welllist.
            date (str): Date when the welllist is defined. Persists until the next date is defined.
        """
        super().__init__(name=name, wells=wells, date=date)

    @staticmethod
    def get_keyword_mapping() -> dict[str, tuple[str, type]]:
        return {}

    @property
    def units(self) -> BaseUnitMapping:
        """Returns the attribute to unit map for the data object."""
        return BaseUnitMapping(None)



"""Class for collection of Nexus water property methods."""
from __future__ import annotations
from dataclasses import dataclass
import os
from typing import Optional, MutableMapping
from ResSimpy.Enums.UnitsEnum import UnitSystem
from ResSimpy.Nexus.DataModels.NexusFile import NexusFile
from ResSimpy.Nexus.DataModels.NexusWaterMethod import NexusWaterMethod
from ResSimpy.Water import Water


@dataclass(kw_only=True)
class NexusWaterMethods(Water):
    """Class for collection of Nexus water property methods.

    Attributes:
        inputs (dict[int, NexusWaterMethod]): Collection of Nexus water property methods, as a dictionary
        files (dict[int, NexusFile]): Dictionary collection of water property files, as defined in Nexus fcs file.
    """

    __inputs: MutableMapping[int, NexusWaterMethod]
    __files: dict[int, NexusFile]
    __properties_loaded: bool = False  # Used in lazy loading
    __model_unit_system: UnitSystem

    def __init__(self, model_unit_system: UnitSystem,
                 inputs: Optional[MutableMapping[int, NexusWaterMethod]] = None,
                 files: Optional[dict[int, NexusFile]] = None) -> None:
        """Initialises the NexusWaterMethods class.

        Args:
            model_unit_system (UnitSystem): Unit system used in the model.
            inputs (Optional[MutableMapping[int, NexusWaterMethod]]): Collection of Nexus water property methods.
            files (Optional[dict[int, NexusFile]]): Collection of water property files, as defined in Nexus fcs file.
                Keyed by the method number.
        """
        if inputs:
            self.__inputs = inputs
        else:
            self.__inputs: MutableMapping[int, NexusWaterMethod] = {}
        if files:
            self.__files = files
        else:
            self.__files = {}
        self.__model_unit_system = model_unit_system
        super().__init__()

    def __repr__(self) -> str:
        """Pretty printing water methods."""
        if not self.__properties_loaded:
            self.load_water_methods()
        printable_str = ''
        for table_num in self.__inputs.keys():
            printable_str += '\n--------------------------------\n'
            printable_str += f'WATER method {table_num}\n'
            printable_str += '--------------------------------\n'
            printable_str += self.__inputs[table_num].__repr__()
            printable_str += '\n'

        return printable_str

    @property
    def inputs(self) -> MutableMapping[int, NexusWaterMethod]:
        if not self.__properties_loaded:
            self.load_water_methods()
        return self.__inputs

    @property
    def files(self) -> dict[int, NexusFile]:
        return self.__files

    def load_water_methods(self):
        # Read in water properties from Nexus water method files
        if self.__files is not None and len(self.__files) > 0:  # Check if water files exist
            for table_num in self.__files.keys():  # For each water property method
                water_file = self.__files[table_num]
                if water_file.location is None:
                    raise ValueError(f'Unable to find water file: {water_file}')
                if os.path.isfile(water_file.location):
                    # Create NexusWaterMethod object
                    self.__inputs[table_num] = NexusWaterMethod(file=water_file, input_number=table_num,
                                                                model_unit_system=self.__model_unit_system)
                    self.__inputs[table_num].read_properties()  # Populate object with water properties in file
        self.__properties_loaded = True

    @property
    def model_unit_system(self) -> UnitSystem:
        """Return the model unit system."""
        return self.__model_unit_system



"""The abstract base class for all wells."""
from __future__ import annotations
from abc import ABC
from dataclasses import dataclass
from typing import Optional, Sequence, Union

from ResSimpy.Completion import Completion
from ResSimpy.Enums.UnitsEnum import UnitSystem
from ResSimpy.Enums.WellTypeEnum import WellType


@dataclass
class Well(ABC):
    _completions: list[Completion]
    _well_name: str
    __unit_system: UnitSystem
    _well_type: Optional[WellType]

    def __init__(self, well_name: str, completions: list[Completion], unit_system: UnitSystem,
                 well_type: Optional[WellType] = None) -> None:
        """Initialises the ConstraintUnits class.

        Args:
            well_name (str): The name of the well.
            completions (list[Completion]): A list of all of the completions on the well.
            unit_system (None | UnitSystem): The unit system associated with the properties on this well.
            well_type (Optional[WellType]): The type of the well represented by a WellType Enum.
        """
        self._well_name = well_name
        self._completions = completions
        self.__unit_system = unit_system
        self._well_type = well_type

    @property
    def completions(self) -> list[Completion]:
        """A list of all of the completions on the well."""
        return self._completions

    @property
    def well_name(self) -> str:
        """The well name."""
        return self._well_name

    @property
    def unit_system(self) -> UnitSystem:
        """The unit system associated with the properties on this well."""
        return self.__unit_system

    @property
    def well_type(self) -> WellType | None:
        """The type of the well."""
        return self._well_type

    @well_type.setter
    def well_type(self, val: WellType) -> None:
        """Sets the well type."""
        if not isinstance(val, WellType):
            raise ValueError(f"Invalid well type: {val}")

        self._well_type = val

    @property
    def dates_of_completions(self) -> list[str]:
        """Returns a list of dates that the well was changed using a completion."""

        dates_changed: list[str] = []
        for completion in self._completions:
            if completion.date not in dates_changed:
                dates_changed.append(completion.date)

        return dates_changed

    @property
    def perforations(self) -> list[Completion]:
        """Returns a list of all of the perforations for the well."""

        activations = filter(lambda x: x.completion_is_perforation, self._completions)
        return list(activations)

    @property
    def first_perforation(self) -> Optional[Completion]:
        """Returns the first perforation for the well."""
        if len(self.perforations) == 0:
            return None

        return self.perforations[0]

    @property
    def shutins(self) -> Sequence[Completion]:
        """Returns a list of all of the shut-ins for the well."""

        shutins = filter(lambda x: x.completion_is_shutin, self._completions)
        return list(shutins)

    @property
    def last_shutin(self) -> Optional[Completion]:
        """Returns the last shut-in for the well in the Wellspec file."""
        if len(self.shutins) == 0:
            return None

        return self.shutins[-1]

    @property
    def printable_well_info(self) -> str:
        """Returns some printable well information in string format."""
        printable_dates_of_completions = ", ".join(self.dates_of_completions)
        well_info = \
            f"""
    Well Name: {self.well_name}
    First Perforation: {'N/A' if self.first_perforation is None else self.first_perforation.date}
    Last Shut-in: {'N/A' if self.last_shutin is None else self.last_shutin.date}
    Dates Changed: {'N/A' if len(self.dates_of_completions) == 0 else printable_dates_of_completions}
    """

        return well_info

    @property
    def open_and_shut_events(self) -> list[tuple[str, Union[int, tuple[float, float]]]]:
        """Returns a list of dates and values representing either the layer, or the depths of each perforation."""
        events = []
        using_k_values: Optional[bool] = None

        for completion in self._completions:
            is_perforation = completion.completion_is_perforation
            if not is_perforation:
                continue
            if completion.k is not None and using_k_values is not False:
                using_k_values = True
                events.append((completion.date, completion.k))
            elif completion.depth_to_top is not None and using_k_values is not True:
                using_k_values = False
                events.append((completion.date, (completion.depth_to_top, completion.depth_to_bottom)))

        return events



from __future__ import annotations
from abc import ABC, abstractmethod
from dataclasses import dataclass
import pandas as pd
from typing import Optional, Sequence, TYPE_CHECKING
from ResSimpy.GridArrayFunction import GridArrayFunction

if TYPE_CHECKING:
    from ResSimpy.File import File


@dataclass
class GridArrayDefinition:
    """Initialises the NexusGrid class.

    Args:
        modifier (Optional[str]): the modifier for the grid array property (e.g. CON, MULT, etc.)
        value (Optional[str]): the actual values for the grid array property in question. Can be an include file.
        mods (Optional[dict[str, pd.DataFrame]): if the grid array has an associated mod card we capture it.
        keyword_in_include_file (bool): an indicator to tell you if the grid array keyword was found in an inc file
    """
    modifier: Optional[str] = None
    value: Optional[str] = None
    # need a parameter for MOD cards
    mods: Optional[dict[str, pd.DataFrame]] = None
    # make a boolean to indicate if a keyword is found in an include file
    # assume initially that it is not
    keyword_in_include_file: bool = False


@dataclass(kw_only=True)
class Grid(ABC):
    # Grid dimensions
    _range_x: Optional[int]
    _range_y: Optional[int]
    _range_z: Optional[int]

    _netgrs: GridArrayDefinition
    _porosity: GridArrayDefinition
    _sw: GridArrayDefinition
    _sg: GridArrayDefinition
    _pressure: GridArrayDefinition
    _temperature: GridArrayDefinition
    _kx: GridArrayDefinition
    _ky: GridArrayDefinition
    _kz: GridArrayDefinition
    _grid_properties_loaded: bool = False

    def __init__(self, assume_loaded: bool = False) -> None:
        """Initialises the Grid class."""
        self._netgrs = GridArrayDefinition()
        self._porosity = GridArrayDefinition()
        self._sw = GridArrayDefinition()
        self._sg = GridArrayDefinition()
        self._pressure = GridArrayDefinition()
        self._temperature = GridArrayDefinition()
        self._kx = GridArrayDefinition()
        self._ky = GridArrayDefinition()
        self._kz = GridArrayDefinition()

        # Grid dimensions
        self._range_x: Optional[int] = None
        self._range_y: Optional[int] = None
        self._range_z: Optional[int] = None
        self._grid_properties_loaded = assume_loaded

    @property
    def range_x(self):
        self.load_grid_properties_if_not_loaded()
        return self._range_x

    @property
    def range_y(self):
        self.load_grid_properties_if_not_loaded()
        return self._range_y

    @property
    def range_z(self):
        self.load_grid_properties_if_not_loaded()
        return self._range_z

    @property
    def netgrs(self):
        self.load_grid_properties_if_not_loaded()
        return self._netgrs

    @property
    def porosity(self):
        self.load_grid_properties_if_not_loaded()
        return self._porosity

    @property
    def sw(self):
        self.load_grid_properties_if_not_loaded()
        return self._sw

    @property
    def sg(self):
        self.load_grid_properties_if_not_loaded()
        return self._sg

    @property
    def pressure(self):
        self.load_grid_properties_if_not_loaded()
        return self._pressure

    @property
    def temperature(self):
        self.load_grid_properties_if_not_loaded()
        return self._temperature

    @property
    def kx(self):
        self.load_grid_properties_if_not_loaded()
        return self._kx

    @property
    def ky(self):
        self.load_grid_properties_if_not_loaded()
        return self._ky

    @property
    def kz(self):
        self.load_grid_properties_if_not_loaded()
        return self._kz

    @abstractmethod
    def load_structured_grid_file(self, structure_grid_file: File, lazy_loading: bool) -> Grid:
        raise NotImplementedError("Implement this in the derived class")

    @abstractmethod
    def load_grid_properties_if_not_loaded(self) -> None:
        raise NotImplementedError("Implement this in the derived class")

    @abstractmethod
    def to_dict(self) -> dict[str, Optional[int] | GridArrayDefinition]:
        raise NotImplementedError("Implement this in the derived class")

    @abstractmethod
    def update_properties_from_dict(self, data: dict[str, int | GridArrayDefinition]) -> None:
        raise NotImplementedError("Implement this in the derived class")

    @property
    def array_functions(self) -> Optional[Sequence[GridArrayFunction]]:
        """Returns a list of the array functions defined in the structured grid file."""
        raise NotImplementedError("Implement this in the derived class")



from dataclasses import dataclass
from abc import ABC

from ResSimpy.DataObjectMixin import DataObjectMixin
from ResSimpy.Enums.FrequencyEnum import FrequencyEnum
from ResSimpy.Enums.OutputType import OutputType
from ResSimpy.Units.AttributeMappings.BaseUnitMapping import BaseUnitMapping


@dataclass(kw_only=True)
class OutputRequest(DataObjectMixin, ABC):
    """Class to hold data input for an Output Request.

    Attributes:
        date (str): Date of the output request.
        output (str): Output request.
        output_type (OutputType): Type of output request.
        output_frequency (FrequencyEnum): Frequency of the output request.
        output_frequency_number (None | float): Number of the frequency of the output request.
    """

    date: str
    output: str
    output_type: OutputType
    output_frequency: FrequencyEnum
    output_frequency_number: None | float

    @staticmethod
    def get_keyword_mapping() -> dict[str, tuple[str, type]]:
        """No keywords for this class, returns an empty dict."""
        return {}

    @property
    def units(self) -> BaseUnitMapping:
        return BaseUnitMapping(unit_system=None)


@dataclass(kw_only=True)
class OutputContents(ABC):
    """Class to hold data input for an Output Contents.

    Attributes:
        date (str): Date of the output contents.
        output (str): Output contents.
        output_type (OutputType): Type of output contents.
    """

    date: str
    output: str
    output_contents: list[str]
    output_type: OutputType



from __future__ import annotations

from abc import ABC
from dataclasses import dataclass
from typing import Sequence, TYPE_CHECKING

from ResSimpy.output_request import OutputRequest, OutputContents
if TYPE_CHECKING:
    from ResSimpy.Simulator import Simulator


@dataclass(kw_only=True)
class Reporting(ABC):
    """Base class for reporting in ResSimpy."""

    __ss_output_requests: Sequence[OutputRequest]
    __ss_output_contents: Sequence[OutputContents]
    __array_output_requests: Sequence[OutputRequest]
    __array_output_contents: Sequence[OutputContents]

    def __init__(self, model: Simulator) -> None:
        """Initialises the Reporting class with a parent model."""
        self.__model = model
        self.__ss_output_requests = []
        self.__ss_output_contents = []
        self.__array_output_requests = []
        self.__array_output_contents = []

    @property
    def ss_output_requests(self) -> Sequence[OutputRequest]:
        """Gets the spreadsheet and tabulated output requests."""
        return self.__ss_output_requests

    @property
    def ss_output_contents(self) -> Sequence[OutputContents]:
        """Gets the contents requested for spreadsheet and tabulated results."""
        return self.__ss_output_contents

    @property
    def array_output_requests(self) -> Sequence[OutputRequest]:
        """Gets the array output requests."""
        return self.__array_output_requests

    @property
    def array_output_contents(self) -> Sequence[OutputContents]:
        """Gets the contents requested to be output for the arrays."""
        return self.__array_output_contents



"""Enum representing the different timestepping and solver methods."""
from enum import Enum


class TimeSteppingMethod(str, Enum):
    """Enum representing the different timestepping and solver methods."""
    implicit = 'IMPLICIT'
    impes = 'IMPES'



"""Abstract base class for holding the network data for holding all components of the production networks."""

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Optional

from ResSimpy.Constraints import Constraints
from ResSimpy.NodeConnections import NodeConnections
from ResSimpy.Nodes import Nodes
from ResSimpy.Targets import Targets
from ResSimpy.WellConnections import WellConnections
from ResSimpy.WellLists import WellLists
from ResSimpy.Wellbores import Wellbores
from ResSimpy.Wellheads import Wellheads


@dataclass(kw_only=True, init=False)
class Network(ABC):
    """Abstract base class for holding the network data for holding all components of the production networks."""
    nodes: Optional[Nodes]
    connections: Optional[NodeConnections]
    constraints: Optional[Constraints]
    targets: Optional[Targets] = None
    well_connections: Optional[WellConnections] = None
    wellheads: Optional[Wellheads] = None
    wellbores: Optional[Wellbores] = None
    welllists: Optional[WellLists] = None
    __has_been_loaded: bool = False

    @abstractmethod
    def load(self) -> None:
        raise NotImplementedError("Implement this in the derived class")

    def get_load_status(self) -> bool:
        """Checks load status and loads the network if it hasn't already been loaded."""
        if not self.__has_been_loaded:
            self.load()
        return self.__has_been_loaded
